from types import TracebackType
from dataclasses import dataclass
from typing import Any, List, Type, Tuple, Generic, Literal, Optional, Sequence

from betterproto.lib.google.protobuf import Any as BetterAny

from qm.type_hinting import NumberT
from qm.exceptions import QmQuaException
from qm.qua._scope_management.scopes_manager import scopes_manager
from qm.qua._expressions import Scalar, QuaScalar, to_scalar_pb_expression
from qm.qua._scope_management._core_scopes import _LeadingScope, _FollowingScope
from qm.grpc.qua import (
    QuaProgramElseIf,
    QuaProgramIfStatement,
    QuaProgramAnyStatement,
    QuaProgramForStatement,
    QuaProgramForEachStatement,
    QuaProgramVarRefExpression,
    QuaProgramAnyScalarExpression,
    QuaProgramAssignmentStatement,
    QuaProgramStatementsCollection,
    QuaProgramArrayVarRefExpression,
    QuaProgramStrictTimingStatement,
    QuaProgramArbitraryContextStatement,
    QuaProgramForEachStatementVariableWithValues,
)


class _ForScope(_LeadingScope):
    def __init__(
        self,
        init: Optional[QuaProgramAssignmentStatement],
        condition: Optional[QuaProgramAnyScalarExpression],
        update: Optional[QuaProgramAssignmentStatement],
        loc: str,
    ):
        super().__init__(loc)
        self._init = init
        self._condition = condition
        self._update = update

    def _create_statement(self) -> QuaProgramAnyStatement:
        init = [QuaProgramAnyStatement(assign=self._init)] if self._init else []
        update = [QuaProgramAnyStatement(assign=self._update)] if self._update else []
        statement = QuaProgramForStatement(
            init=QuaProgramStatementsCollection(statements=init),
            update=QuaProgramStatementsCollection(statements=update),
            body=QuaProgramStatementsCollection(statements=self._statements),
        )
        if self._condition:
            statement.condition = self._condition
        return QuaProgramAnyStatement(for_=statement)


class _ForEachScope(_LeadingScope):
    def __init__(
        self,
        iterators: Sequence[Tuple[QuaProgramVarRefExpression, QuaProgramArrayVarRefExpression]],
        loc: str,
    ) -> None:
        super().__init__(loc)
        self._iterators = iterators

    def _create_statement(self) -> QuaProgramAnyStatement:
        statement = QuaProgramForEachStatement(
            iterator=[
                QuaProgramForEachStatementVariableWithValues(variable=var, array=arr) for var, arr in self._iterators
            ],
            body=QuaProgramStatementsCollection(statements=self._statements),
            loc=self._loc,
        )
        return QuaProgramAnyStatement(for_each=statement)


class _IfScope(_LeadingScope):
    def __init__(self, condition: Scalar[bool], unsafe: bool, loc: str):
        super().__init__(loc)
        self._condition = condition
        self._unsafe = unsafe

    def _create_statement(self) -> QuaProgramAnyStatement:
        statement = QuaProgramIfStatement(
            condition=to_scalar_pb_expression(self._condition),
            unsafe=self._unsafe,
            body=QuaProgramStatementsCollection(statements=self._statements),
            loc=self._loc,
        )
        return QuaProgramAnyStatement(if_=statement)


class _ElifScope(_FollowingScope):
    def __init__(self, condition: Scalar[bool], if_statement: QuaProgramIfStatement, loc: str):
        super().__init__()
        self._condition = condition
        self._if_statement = if_statement
        self._loc = loc

    def _add_to_leading_scope(self) -> None:
        self._if_statement.elseifs.append(self._create_elif_statement())

    def _create_elif_statement(self) -> QuaProgramElseIf:
        return QuaProgramElseIf(
            loc=self._loc,
            condition=to_scalar_pb_expression(self._condition),
            body=QuaProgramStatementsCollection(statements=self._statements),
        )


class _ElseScope(_FollowingScope):
    def __init__(self, if_statement: QuaProgramIfStatement):
        super().__init__()
        self._if_statement = if_statement

    def _add_to_leading_scope(self) -> None:
        self._if_statement.else_ = QuaProgramStatementsCollection(statements=self._statements)


class _StrictTimingScope(_LeadingScope):
    def _create_statement(self) -> QuaProgramAnyStatement:
        statement = QuaProgramStrictTimingStatement(
            body=QuaProgramStatementsCollection(statements=self._statements),
            loc=self._loc,
        )
        return QuaProgramAnyStatement(strict_timing=statement)


class _SwitchScope(_LeadingScope, Generic[NumberT]):
    """Switch scope does not hold statements"""

    def __init__(self, expression: QuaScalar[NumberT], unsafe: bool, loc: str):
        super().__init__(loc)
        self._expression: QuaScalar[NumberT] = expression
        self._unsafe = unsafe
        self.cases: List[_CaseData[NumberT]] = []
        self.default: List[QuaProgramAnyStatement] = []

    def _create_statement(self) -> QuaProgramAnyStatement:
        if not self.cases:
            raise QmQuaException(
                "Expecting scope with body. Expecting switch scope with body. "
                "This switch scope does not have any cases"
            )

        conditions = [(self._expression.__eq__(case.value)).unwrapped for case in self.cases]

        first_case = self.cases[0]
        first_condition = conditions[0]

        else_ifs = [
            QuaProgramElseIf(
                condition=cond,
                body=QuaProgramStatementsCollection(statements=case.statements),
                loc=self._loc,
            )
            for cond, case in zip(conditions[1:], self.cases[1:])
        ]
        statement = QuaProgramIfStatement(
            condition=first_condition,
            body=QuaProgramStatementsCollection(statements=first_case.statements),
            unsafe=self._unsafe,
            elseifs=else_ifs,
            else_=QuaProgramStatementsCollection(statements=self.default),
            loc=self._loc,
        )
        return QuaProgramAnyStatement(if_=statement)

    def append_statement(self, statement: QuaProgramAnyStatement) -> None:
        raise QmQuaException(
            "Switch scope can only contain 'case' and 'default' blocks. Direct statements are not allowed (Expecting scope with body.)."
        )


@dataclass
class _CaseData(Generic[NumberT]):
    value: QuaScalar[NumberT]
    statements: List[QuaProgramAnyStatement]


class _CaseScope(_FollowingScope, Generic[NumberT]):
    def __init__(self, switch_scope: "_SwitchScope[NumberT]", value: QuaScalar[NumberT]):
        super().__init__()
        self._switch_scope: "_SwitchScope[NumberT]" = switch_scope
        self._value: QuaScalar[NumberT] = value

    def _add_to_leading_scope(self) -> None:
        data = _CaseData(value=self._value, statements=self._statements)
        self._switch_scope.cases.append(data)


class _CaseDefaultScope(_FollowingScope):
    def __init__(self, switch_scope: "_SwitchScope[Any]"):
        super().__init__()
        self._switch_scope: "_SwitchScope[Any]" = switch_scope

    def _add_to_leading_scope(self) -> None:
        self._switch_scope.default = self._statements


class _PortConditionScope:
    """
    Not a "real" scope, since it doesn't inherit from _BaseScope and therefore is not added to the scopes stack.
    Unlike other scopes where the scopes_manager accesses them through the scopes stack, this scope updates the
    scope manager through dedicated 'set' functions.
    """

    def __init__(self, expression: QuaProgramAnyScalarExpression):
        self._expression = expression

    def __enter__(self) -> None:
        scopes_manager._set_port_condition(self._expression)

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> Literal[False]:
        scopes_manager._unset_port_condition()
        return False


class _ArbitraryScope(_LeadingScope):
    def __init__(
        self,
        loc: str,
        name: str,
        data: Optional[BetterAny],
    ):
        super().__init__(loc)
        self._name = name
        self._data = data

    def _create_statement(self) -> QuaProgramAnyStatement:
        statement = QuaProgramArbitraryContextStatement(
            loc=self._loc,
            name=self._name,
            data=self._data,
            body=QuaProgramStatementsCollection(statements=self._statements),
        )
        return QuaProgramAnyStatement(arbitrary_context=statement)



import uuid
import warnings
from typing import Dict, Union, Optional

import betterproto

from qm.utils import deprecation_message
from qm.api.models.capabilities import OPX_FEM_IDX, QopCaps, ServerCapabilities
from qm.program._dict_to_pb_converter.base_converter import BaseDictToPbConverter
from qm.program._dict_to_pb_converter.converters.pulse_converter import PulseConverter
from qm.program._dict_to_pb_converter.converters.octave_converter import OctaveConverter
from qm.program._dict_to_pb_converter.converters.waveform_converter import WaveformConverter
from qm.program._dict_to_pb_converter.converters.oscillator_converter import OscillatorConverter
from qm.utils.config_utils import get_logical_pb_config, get_fem_config_instance, get_controller_pb_config
from qm.program._dict_to_pb_converter.converters.mixer_correction_converter import MixerCorrectionConverter
from qm.program._dict_to_pb_converter.converters.integration_weights_converter import IntegrationWeightsConverter
from qm.program._dict_to_pb_converter.converters.element_converter import ElementConverter, StickyElementIsNotSupported
from qm.type_hinting.config_types import FullQuaConfig, LogicalQuaConfig, ControllerQuaConfig, DigitalWaveformConfigType
from qm.program._dict_to_pb_converter.converters.control_device_converter.control_device_converter import (
    ControlDeviceConverter,
)
from qm.exceptions import (
    ConfigValidationException,
    OctaveUnsupportedOnUpdate,
    ConfigurationLockedByOctave,
    CapabilitiesNotInitializedError,
    ElementInputConnectionAmbiguity,
    ElementOutputConnectionAmbiguity,
)
from qm.grpc.qua_config import (
    QuaConfig,
    QuaConfigMatrix,
    QuaConfigMixerDec,
    QuaConfigDeviceDec,
    QuaConfigMixInputs,
    QuaConfigElementDec,
    QuaConfigQuaConfigV1,
    QuaConfigQuaConfigV2,
    QuaConfigOctaveConfig,
    QuaConfigControllerDec,
    QuaConfigLogicalConfig,
    QuaConfigCorrectionEntry,
    QuaConfigMultipleOutputs,
    QuaConfigControllerConfig,
    QuaConfigDigitalWaveformDec,
    QuaConfigOctaveRfOutputConfig,
    QuaConfigDigitalWaveformSample,
    QuaConfigMicrowaveOutputPortReference,
)


class DictToQuaConfigConverter(
    BaseDictToPbConverter[Union[FullQuaConfig, ControllerQuaConfig, LogicalQuaConfig], QuaConfig]
):
    def __init__(
        self, capabilities: ServerCapabilities, init_mode: bool = True, octave_already_configured: bool = False
    ):
        super().__init__(capabilities, init_mode)
        self.control_device_converter = ControlDeviceConverter(capabilities, init_mode)
        self.octave_converter = OctaveConverter(capabilities, init_mode)
        self.element_converter = ElementConverter(capabilities, init_mode)
        self.pulse_converter = PulseConverter(capabilities, init_mode)
        self.waveform_converter = WaveformConverter(capabilities, init_mode)
        self.iw_converter = IntegrationWeightsConverter(capabilities, init_mode)
        self.mixer_correction_converter = MixerCorrectionConverter(capabilities, init_mode)
        self.oscillator_converter = OscillatorConverter(capabilities, init_mode)

        self.octave_already_configured = octave_already_configured

    def convert(
        self,
        input_data: Union[FullQuaConfig, ControllerQuaConfig, LogicalQuaConfig],
    ) -> QuaConfig:
        self.run_preload_validations(input_data, self.octave_already_configured)

        pb_config = self.set_config_wrapper()
        controller_config = get_controller_pb_config(pb_config)
        logical_config = get_logical_pb_config(pb_config)

        def set_controllers() -> None:
            for k, v in input_data["controllers"].items():  # type: ignore[typeddict-item]
                controller_config.control_devices[k] = self.control_device_converter.convert(v)
            # Controllers attribute is supported only in config v1
            if self.all_controllers_are_opx(controller_config.control_devices) and isinstance(
                controller_config, QuaConfigQuaConfigV1
            ):
                for _k, _v in controller_config.control_devices.items():
                    controller_inst = get_fem_config_instance(_v.fems[OPX_FEM_IDX])
                    if not isinstance(controller_inst, QuaConfigControllerDec):
                        raise ValueError("This should not happen")
                    controller_config.controllers[_k] = controller_inst

        def set_octaves() -> None:
            for k, v in input_data.get("octaves", {}).items():  # type: ignore[attr-defined]
                controller_config.octaves[k] = self.octave_converter.convert(v)

        def set_elements() -> None:
            for k, v in input_data["elements"].items():  # type: ignore[typeddict-item]
                try:
                    logical_config.elements[k] = self.element_converter.convert(v)
                except StickyElementIsNotSupported:
                    raise ConfigValidationException(f"Server does not support digital sticky used in element " f"'{k}'")

        def set_pulses() -> None:
            for k, v in input_data["pulses"].items():  # type: ignore[typeddict-item]
                logical_config.pulses[k] = self.pulse_converter.convert(v)

        def set_waveforms() -> None:
            for k, v in input_data["waveforms"].items():  # type: ignore[typeddict-item]
                logical_config.waveforms[k] = self.waveform_converter.convert(v)

        def set_digital_waveforms() -> None:
            for k, v in input_data["digital_waveforms"].items():  # type: ignore[typeddict-item]
                logical_config.digital_waveforms[k] = QuaConfigDigitalWaveformDec(
                    samples=[QuaConfigDigitalWaveformSample(value=bool(s[0]), length=s[1]) for s in v["samples"]]
                )

        def set_integration_weights() -> None:
            for k, v in input_data["integration_weights"].items():  # type: ignore[typeddict-item]
                logical_config.integration_weights[k] = self.iw_converter.convert(v)

        def set_mixers() -> None:
            for k, v in input_data["mixers"].items():  # type: ignore[typeddict-item]
                controller_config.mixers[k] = QuaConfigMixerDec(
                    correction=[self.mixer_correction_converter.convert(u) for u in v]
                )

        def set_oscillators() -> None:
            for k, v in input_data["oscillators"].items():  # type: ignore[typeddict-item]
                logical_config.oscillators[k] = self.oscillator_converter.convert(v)

        key_to_action = {
            "version": lambda: None,
            "controllers": set_controllers,
            "elements": set_elements,
            "pulses": set_pulses,
            "waveforms": set_waveforms,
            "digital_waveforms": set_digital_waveforms,
            "integration_weights": set_integration_weights,
            "mixers": set_mixers,
            "oscillators": set_oscillators,
            "octaves": set_octaves,
        }

        if "version" in input_data:
            warnings.warn(
                deprecation_message("version", "1.2.2", "2.0.0", "Please remove it from the QUA config."),
                DeprecationWarning,
            )

        for key in input_data:
            key_to_action[key]()

        self.apply_post_load_setters(pb_config)

        return pb_config

    def run_preload_validations(
        self,
        config: Union[FullQuaConfig, ControllerQuaConfig, LogicalQuaConfig],
        octave_already_configured: bool = False,
    ) -> None:
        # When the capabilities aren't initialized, the capabilities argument is of type 'Provide' instead of 'ServerCapabilities'.
        # This is only relevant for qua_config_schema.py, where the capabilities are provided by the container.
        if not isinstance(self._capabilities, ServerCapabilities):
            raise CapabilitiesNotInitializedError

        if not self._init_mode:
            # With these two validations, we ensure any configuration that relates to Octave is done in init mode.
            # Or in other words, Octave doesn't support 'send program with config'.

            if "octaves" in config:
                raise OctaveUnsupportedOnUpdate("Octaves are not supported in non-init mode")

            logical_config_present = any(key in config for key in LogicalQuaConfig.__annotations__)
            if octave_already_configured and logical_config_present:
                # If Octaves were already configured, we cannot change the logical configuration anymore, because it may override
                # automatic configurations that were done for Octave, like the ones in "apply_post_load_setters()".
                raise ConfigurationLockedByOctave(
                    "Since Octaves were used in the initial configuration, no modifications to the logical configuration are allowed — whether related to Octaves or not. "
                    "To resolve this, either avoid using Octaves, or ensure all logical configuration is completed when opening the QM."
                )

    def set_config_wrapper(self) -> QuaConfig:
        pb_config = QuaConfig()

        if self._capabilities.supports(QopCaps.config_v2):
            pb_config.v2 = QuaConfigQuaConfigV2(
                controller_config=QuaConfigControllerConfig(), logical_config=QuaConfigLogicalConfig()
            )
        else:
            pb_config.v1_beta = QuaConfigQuaConfigV1()

        return pb_config

    @staticmethod
    def all_controllers_are_opx(control_devices: Dict[str, QuaConfigDeviceDec]) -> bool:
        for device_config in control_devices.values():
            for fem_config in device_config.fems.values():
                _, controller_inst = betterproto.which_one_of(fem_config, "fem_type_one_of")
                if not isinstance(controller_inst, QuaConfigControllerDec):
                    return False
        return True

    def apply_post_load_setters(self, pb_config: QuaConfig) -> None:
        # In config_v2, elements can be defined independently of mixers.
        # This breaks the existing logic, which automatically assigns default mixers and octave values based on the elements.
        # As a result, users of config_v2 that rely on these autocompleted values need to manually add them, or use the old flow (sending full config in init mode).
        # The long-term goal is to move autocompletion logic into the gateway. For more details, see: https://quantum-machines.atlassian.net/browse/OPXK-25086
        if self._init_mode:
            self.set_octave_upconverter_connection_to_elements(pb_config)
            self.set_lo_frequency_to_mix_input_elements_that_are_connected_to_octave(pb_config)
            self.set_octave_downconverter_connection_to_elements(pb_config)
            self.set_non_existing_mixers_in_mix_input_elements(pb_config)

    @staticmethod
    def set_octave_upconverter_connection_to_elements(pb_config: QuaConfig) -> None:
        octaves_config = get_controller_pb_config(pb_config).octaves
        elements_config = get_logical_pb_config(pb_config).elements

        for element in elements_config.values():
            for rf_input in element.rf_inputs.values():
                if rf_input.device_name in octaves_config:
                    if rf_input.port in octaves_config[rf_input.device_name].rf_outputs:
                        _, element_input = betterproto.which_one_of(element, "element_inputs_one_of")
                        if element_input is not None:
                            raise ElementInputConnectionAmbiguity("Ambiguous definition of element input")

                        upconverter_config = octaves_config[rf_input.device_name].rf_outputs[rf_input.port]
                        element.mix_inputs = QuaConfigMixInputs(
                            i=upconverter_config.i_connection, q=upconverter_config.q_connection
                        )

    def set_lo_frequency_to_mix_input_elements_that_are_connected_to_octave(self, pb_config: QuaConfig) -> None:
        octaves_config = get_controller_pb_config(pb_config).octaves
        elements_config = get_logical_pb_config(pb_config).elements

        for element in elements_config.values():
            _, element_input = betterproto.which_one_of(element, "element_inputs_one_of")
            if isinstance(element_input, QuaConfigMixInputs):
                rf_output = self._get_rf_output_for_octave(element, octaves_config)
                if rf_output is None:
                    continue

                if element_input.lo_frequency not in {0, int(rf_output.lo_frequency)}:
                    raise ConfigValidationException(
                        "LO frequency mismatch. The frequency stated in the element is different from "
                        "the one stated in the Octave, remove the one in the element."
                    )
                element_input.lo_frequency = int(rf_output.lo_frequency)
                if self._capabilities.supports_double_frequency:
                    element_input.lo_frequency_double = rf_output.lo_frequency

    @staticmethod
    def _get_rf_output_for_octave(
        element: QuaConfigElementDec, octaves: Dict[str, QuaConfigOctaveConfig]
    ) -> Optional[QuaConfigOctaveRfOutputConfig]:
        if element.rf_inputs:
            element_rf_input = list(element.rf_inputs.values())[0]
            octave_config = octaves[element_rf_input.device_name]
            return octave_config.rf_outputs[element_rf_input.port]

        # This part is for users that do not use the rf_inputs  for connecting the octave
        element_input = element.mix_inputs
        for octave in octaves.values():
            for rf_output in octave.rf_outputs.values():
                if all(
                    [
                        (rf_output.i_connection.controller == element_input.i.controller),
                        (rf_output.i_connection.fem == element_input.i.fem),
                        (rf_output.i_connection.number == element_input.i.number),
                        (rf_output.q_connection.controller == element_input.q.controller),
                        (rf_output.q_connection.fem == element_input.q.fem),
                        (rf_output.q_connection.number == element_input.q.number),
                    ]
                ):
                    return rf_output
        return None

    @staticmethod
    def set_octave_downconverter_connection_to_elements(pb_config: QuaConfig) -> None:
        octaves_config = get_controller_pb_config(pb_config).octaves
        elements_config = get_logical_pb_config(pb_config).elements

        for element in elements_config.values():
            for _, rf_output in element.rf_outputs.items():
                if rf_output.device_name in octaves_config:
                    if rf_output.port in octaves_config[rf_output.device_name].rf_inputs:
                        downconverter_config = octaves_config[rf_output.device_name].if_outputs
                        outputs_form_octave = {
                            downconverter_config.if_out1.name: downconverter_config.if_out1.port,
                            downconverter_config.if_out2.name: downconverter_config.if_out2.port,
                        }
                        for k, v in outputs_form_octave.items():
                            if k in element.outputs:
                                if v != element.outputs[k]:
                                    raise ElementOutputConnectionAmbiguity(
                                        f"Output {k} is connected to {element.outputs[k]} but the octave "
                                        f"downconverter is connected to {v}"
                                    )
                            else:
                                element.outputs[k] = v
                                _, element_outputs = betterproto.which_one_of(element, "element_outputs_one_of")
                                if isinstance(element_outputs, QuaConfigMicrowaveOutputPortReference):
                                    raise ConfigValidationException("Cannot connect octave to microwave output")
                                elif isinstance(element_outputs, QuaConfigMultipleOutputs):
                                    element_outputs.port_references[k] = v
                                else:
                                    element.multiple_outputs = QuaConfigMultipleOutputs(port_references={k: v})

    @staticmethod
    def set_non_existing_mixers_in_mix_input_elements(pb_config: QuaConfig) -> None:
        mixers_config = get_controller_pb_config(pb_config).mixers
        elements_config = get_logical_pb_config(pb_config).elements

        for element_name, element in elements_config.items():
            _, element_input = betterproto.which_one_of(element, "element_inputs_one_of")
            if isinstance(element_input, QuaConfigMixInputs):
                if (
                    element.intermediate_frequency
                ):  # This is here because in validation I saw that we can set an element without IF
                    if not element_input.mixer:
                        element_input.mixer = f"{element_name}_mixer_{uuid.uuid4().hex[:3]}"
                        # The uuid is just to make sure the mixer doesn't exist
                    if element_input.mixer not in mixers_config:
                        mixers_config[element_input.mixer] = QuaConfigMixerDec(
                            correction=[
                                QuaConfigCorrectionEntry(
                                    frequency=element.intermediate_frequency,
                                    frequency_negative=element.intermediate_frequency_negative,
                                    frequency_double=element.intermediate_frequency_double,
                                    lo_frequency=element_input.lo_frequency,
                                    lo_frequency_double=element_input.lo_frequency_double,
                                    correction=QuaConfigMatrix(v00=1, v01=0, v10=0, v11=1),
                                )
                            ]
                        )

    def deconvert(self, output_data: QuaConfig) -> FullQuaConfig:
        controller_config = get_controller_pb_config(output_data)
        logical_config = get_logical_pb_config(output_data)

        if controller_config.control_devices:
            controllers = {
                name: self.control_device_converter.deconvert(value)
                for name, value in controller_config.control_devices.items()
            }
        elif isinstance(controller_config, QuaConfigQuaConfigV1) and controller_config.controllers:
            controllers = {
                name: self.control_device_converter._deconvert_controller(value)
                for name, value in controller_config.controllers.items()
            }
        else:
            controllers = {}

        result: FullQuaConfig = {
            "controllers": controllers,
            "oscillators": {
                name: self.oscillator_converter.deconvert(oscillator)
                for name, oscillator in logical_config.oscillators.items()
            },
            "elements": {
                name: self.element_converter.deconvert(elem) for name, elem in logical_config.elements.items()
            },
            "pulses": {name: self.pulse_converter.deconvert(pulse) for name, pulse in logical_config.pulses.items()},
            "waveforms": {name: self.waveform_converter.deconvert(wf) for name, wf in logical_config.waveforms.items()},
            "digital_waveforms": {
                name: self._deconvert_digital_waveforms(wf) for name, wf in logical_config.digital_waveforms.items()
            },
            "integration_weights": {
                name: self.iw_converter.deconvert(iw) for name, iw in logical_config.integration_weights.items()
            },
            "mixers": {
                name: [self.mixer_correction_converter.deconvert(u) for u in value.correction]
                for name, value in controller_config.mixers.items()
            },
        }

        return result

    @staticmethod
    def _deconvert_digital_waveforms(digital_wave_form: QuaConfigDigitalWaveformDec) -> DigitalWaveformConfigType:
        temp_list: list[tuple[int, int]] = []
        for sample in digital_wave_form.samples:
            value = int(bool(sample.value))
            temp_list.append((value, sample.length))

        return {"samples": temp_list}



from typing import Type, TypeVar

import betterproto
from google.protobuf.any_pb2 import Any as PbAny
from google.protobuf.message import Message as PbMessage
from betterproto.lib.google.protobuf import Any as BetterAny

BetterProtoMessageT = TypeVar("BetterProtoMessageT", bound=betterproto.Message)
ProtoMessageT = TypeVar("ProtoMessageT", bound=PbMessage)


def betterproto_to_any(message: betterproto.Message) -> BetterAny:
    """
    Converts a betterproto message to a betterproto Any message.
    Args:
        message (betterproto.Message): A betterproto message instance.
    Returns:
        BetterAny: The packed Any message.
    """
    any_message = BetterAny()
    any_message.type_url = f"type.googleapis.com/{message.__class__.__name__}"
    any_message.value = message.SerializeToString()
    return any_message


def any_to_betterproto(any_message: BetterAny, message_cls: Type[BetterProtoMessageT]) -> BetterProtoMessageT:
    """
    Converts a protobuf Any message to a betterproto message of the given class.

    Args:
        any_message (any_pb2.Any): The protobuf Any message to unpack.
        message_cls (type[betterproto.Message]): The betterproto message class to unpack into.

    Returns:
        betterproto.Message: The unpacked betterproto message.
    """
    message = message_cls().parse(any_message.value)
    return message


def proto_to_any(message: PbMessage) -> BetterAny:
    """
    Packs a standard protobuf message into a betterproto Any message.

    Args:
        message (PbMessage): A protobuf message instance.

    Returns:
        BetterAny: The packed Any message.
    """
    any_pb_message = PbAny()
    any_pb_message.Pack(message)
    return BetterAny(type_url=any_pb_message.type_url, value=any_pb_message.value)


def any_to_proto(any_message: BetterAny, message_cls: Type[ProtoMessageT]) -> ProtoMessageT:
    """
    Converts a betterproto Any message into a standard protobuf message.

    Args:
        any_message (BetterAny): The betterproto Any message.
        message_cls (Type[ProtoMessageT]): The expected protobuf message class to unpack into.

    Returns:
        PbMessage: The unpacked protobuf message.
    """
    message = message_cls()
    message.ParseFromString(any_message.value)
    return message



import betterproto

from qm.type_hinting.config_types import OscillatorConfigType
from qm.grpc.qua_config import QuaConfigMixerRef, QuaConfigOscillator
from qm.program._dict_to_pb_converter.base_converter import BaseDictToPbConverter


class OscillatorConverter(BaseDictToPbConverter[OscillatorConfigType, QuaConfigOscillator]):
    def convert(self, input_data: OscillatorConfigType) -> QuaConfigOscillator:
        return self.oscillator_to_pb(input_data)

    def oscillator_to_pb(self, data: OscillatorConfigType) -> QuaConfigOscillator:
        oscillator = QuaConfigOscillator()
        if "intermediate_frequency" in data:
            oscillator.intermediate_frequency = int(data["intermediate_frequency"])
            if self._capabilities.supports_double_frequency:
                oscillator.intermediate_frequency_double = float(data["intermediate_frequency"])

        if "mixer" in data:
            oscillator.mixer = QuaConfigMixerRef(mixer=data["mixer"])
            oscillator.mixer.lo_frequency = int(data.get("lo_frequency", 0))
            if self._capabilities.supports_double_frequency:
                oscillator.mixer.lo_frequency_double = float(data.get("lo_frequency", 0.0))

        return oscillator

    def deconvert(self, output_data: QuaConfigOscillator) -> OscillatorConfigType:
        oscillator_config_data: OscillatorConfigType = {}
        if output_data.intermediate_frequency_double:
            freq = output_data.intermediate_frequency_double
            oscillator_config_data["intermediate_frequency"] = freq
        elif output_data.intermediate_frequency:
            freq = int(output_data.intermediate_frequency)
            oscillator_config_data["intermediate_frequency"] = freq
        if betterproto.serialized_on_wire(output_data.mixer):
            if output_data.mixer.mixer:
                oscillator_config_data["mixer"] = output_data.mixer.mixer
            if output_data.mixer.lo_frequency_double:
                lo_freq = output_data.mixer.lo_frequency_double
                oscillator_config_data["lo_frequency"] = float(lo_freq)
            elif output_data.mixer.lo_frequency:
                lo_freq = output_data.mixer.lo_frequency
                oscillator_config_data["lo_frequency"] = int(lo_freq)
        return oscillator_config_data



from typing import Union

from qm.qua._dsl import function_expressions
from qm.api.models.capabilities import QopCaps
from qm.qua._scope_management.scopes_manager import scopes_manager
from qm.qua._expressions import Scalar, QuaBroadcast, QuaArrayVariable, QuaFunctionOutput


def _create_broadcast_expression(value: QuaFunctionOutput[bool]) -> QuaBroadcast[bool]:
    scopes_manager.program_scope.add_used_capability(QopCaps.broadcast)
    return QuaBroadcast(bool, value.unwrapped)


class broadcast:
    """
    A module that provides functionality for creating broadcast expressions.
    Broadcasting allows more control over making a locally measured variable available to all elements in the QUA program.
    """

    @staticmethod
    def and_(*values: Union[Scalar[bool], QuaArrayVariable[bool]]) -> QuaBroadcast[bool]:
        """
        Preforms a logical AND operation on the input values and broadcasts the result to all elements in the QUA program.

        Args:
            *values (boolean, QUA boolean, Qua array of type boolean): The input values to be combined using a
                logical AND operation. Each input can be a single boolean, a QUA boolean or a QUA array of booleans.

        Returns:
            A boolean broadcast object, that can be used as input for any QUA command requiring a QUA boolean.
        """
        return _create_broadcast_expression(function_expressions.and_(*values))

    @staticmethod
    def or_(*values: Union[Scalar[bool], QuaArrayVariable[bool]]) -> QuaBroadcast[bool]:
        """
        Preforms a logical OR operation on the input values and broadcasts the result to all elements in the QUA program.

        Args:
            *values (boolean, QUA boolean, Qua array of type boolean): The input values to be combined using a
                logical OR operation. Each input can be a single boolean, a QUA boolean or a QUA array of booleans.

        Returns:
            A boolean broadcast object, that can be used as input for any QUA command requiring a QUA boolean.
        """
        return _create_broadcast_expression(function_expressions.or_(*values))

    @staticmethod
    def xor_(*values: Union[Scalar[bool], QuaArrayVariable[bool]]) -> QuaBroadcast[bool]:
        """
        Preforms a logical XOR (exclusive OR) operation on the input values and broadcasts the result to all elements in
        the QUA program.

        Args:
            *values (boolean, QUA boolean, Qua array of type boolean): The input values to be combined using a
                logical XOR operation. Each input can be a single boolean, a QUA boolean or a QUA array of booleans.

        Returns:
            A boolean broadcast object, that can be used as input for any QUA command requiring a QUA boolean.
        """
        return _create_broadcast_expression(function_expressions.xor_(*values))



import random
import warnings
from typing import Any, List, Type, Tuple, Union, TypeVar, Callable, Optional, Sequence, overload

from qm._loc import _get_loc
from qm.type_hinting.general import NumberT
from qm.qua._dsl.variable_handling import assign, declare
from qm.utils import deprecation_message, get_iterable_elements_datatype
from qm.grpc.qua import (
    QuaProgramAnyScalarExpression,
    QuaProgramLibFunctionExpression,
    QuaProgramLibFunctionExpressionArgument,
)
from qm.qua._expressions import (
    Scalar,
    Vector,
    QuaVariable,
    ScalarOfAnyType,
    VectorOfAnyType,
    QuaArrayVariable,
    QuaLibFunctionOutput,
    fixed,
    create_qua_scalar_expression,
)

S = TypeVar("S", bool, int, float)
SomeCallable = Callable[..., Any]


def _get_func_lib_and_name(function: SomeCallable) -> Tuple[str, str]:
    lib_name, func_name = function.__qualname__.split(".")
    return lib_name.lower(), func_name


def _create_qua_vector_expression(value: Vector[NumberT]) -> QuaArrayVariable[NumberT]:
    if isinstance(value, QuaArrayVariable):
        return value
    data_type = get_iterable_elements_datatype(value)
    return declare(data_type, value=value)


def _standardize_args(*args: Union[ScalarOfAnyType, VectorOfAnyType]) -> List[QuaProgramLibFunctionExpressionArgument]:
    standardized_args = []

    for arg in args:
        # Checking if the argument is a vector, unfortunately python doesn't allow to use isinstance(arg, Vector)
        if isinstance(arg, (Sequence, QuaArrayVariable)):
            if isinstance(arg, Sequence):
                arg = _create_qua_vector_expression(arg)
            standardized_args.append(QuaProgramLibFunctionExpressionArgument(array=arg.unwrapped))
        else:
            arg = create_qua_scalar_expression(arg)
            standardized_args.append(QuaProgramLibFunctionExpressionArgument(scalar=arg.unwrapped))

    return standardized_args


def __create_output_expression(
    function: SomeCallable,
    output_type: Type[NumberT],
    *args: Union[ScalarOfAnyType, VectorOfAnyType],
) -> QuaLibFunctionOutput[NumberT]:
    standardized_args = _standardize_args(*args)
    lib_name, func_name = _get_func_lib_and_name(function)
    any_scalar_expression = QuaProgramAnyScalarExpression(
        lib_function=QuaProgramLibFunctionExpression(
            function_name=func_name, arguments=standardized_args, library_name=lib_name, loc=_get_loc()
        )
    )
    return QuaLibFunctionOutput(any_scalar_expression, output_type)


def _create_output_expression(
    function: SomeCallable, output_type: Type[NumberT], *args: ScalarOfAnyType
) -> QuaLibFunctionOutput[NumberT]:
    return __create_output_expression(function, output_type, *args)


def _create_vectors_output_expression(
    function: SomeCallable, output_type: Type[S], *args: VectorOfAnyType
) -> QuaLibFunctionOutput[S]:
    return __create_output_expression(function, output_type, *args)


def call_library_function(
    function: SomeCallable, output_type: Type[NumberT], *args: ScalarOfAnyType
) -> QuaLibFunctionOutput[NumberT]:
    warnings.warn(
        deprecation_message(
            method="call_library_function",
            deprecated_in="1.2.2",
            removed_in="2.0.0",
            details="""
Please call the required function directly from the available classes.
For instance, instead of using call_library_function(Random.rand_int, int, x, y), simply use Random(x).rand_int(y).""",
        )
    )
    return _create_output_expression(function, output_type, *args)


def call_vectors_library_function(
    function: SomeCallable, output_type: Type[S], *args: Vector[NumberT]
) -> QuaLibFunctionOutput[S]:
    warnings.warn(
        deprecation_message(
            method="call_vectors_library_function",
            deprecated_in="1.2.2",
            removed_in="2.0.0",
            details="""
Please call the required function directly from the available classes.
For instance, instead of using call_vectors_library_function(Math.sum, x.dtype, x), simply use Math.sum(x).""",
        )
    )
    return _create_vectors_output_expression(function, output_type, *args)


class Math:
    @staticmethod
    def log(x: Scalar[float], base: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes $\mathrm{log}_{base}(x)$

        Args:
            x: a QUA fixed larger than pow2(-8)=0.00390625
            base: a QUA fixed larger than pow2(1/8)=1.09051

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.log, float, x, base)

    @staticmethod
    def pow(base: Scalar[float], x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes ${base}^{x}$.
        Does not support base=1, nor the case where both base=0 & x=0.

        Args:
            base: a non-negative QUA fixed
            x: a QUA fixed

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.pow, float, base, x)

    @staticmethod
    def div(x: Scalar[NumberT], y: Scalar[NumberT]) -> QuaLibFunctionOutput[float]:
        r"""Computes the division between two same-type variables $x/y$.
        Takes integer and fixed variables as long as both are of the same type.
        When both inputs are integer, the return type can be either a fixed point number or an integer, in which case $floor(x/y)$ is outputted.

        Args:
            x: a QUA int or fixed
            y: a QUA int or fixed

        Returns:
            a QUA int or fixed
        """
        return _create_output_expression(Math.div, float, x, y)

    @staticmethod
    def exp(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes $e^{x}$

        Args:
            x: a QUA fixed smaller than ln(8)=2.0794415416

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.exp, float, x)

    @staticmethod
    def pow2(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes $2^{x}$

        Args:
            x: a QUA fixed smaller than 3 (to avoid overflow)

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.pow2, float, x)

    @staticmethod
    def ln(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes $\mathrm{ln}(x)$

        Args:
            x: a QUA fixed larger than exp(-8)=0.0003354627

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.ln, float, x)

    @staticmethod
    def log2(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes $\mathrm{log}_{2}(x)$

        Args:
            x: a QUA fixed larger than pow2(-8)=0.00390625

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.log2, float, x)

    @staticmethod
    def log10(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes $\mathrm{log}_{10}(x)$

        Args:
            x: a QUA fixed larger than pow10(-8)=0.00000001

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.log10, float, x)

    @staticmethod
    def sqrt(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes the square root of x

        Args:
            x: a non-negative QUA fixed

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.sqrt, float, x)

    @staticmethod
    def inv_sqrt(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes the inverse square root of x

        Args:
            x: a QUA fixed larger than 1/64

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.sqrt, float, x)

    @staticmethod
    def inv(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes the inverse of x

        Args:
            x: a QUA fixed which is x<=-1/8 or 1/8<x

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.inv, float, x)

    @staticmethod
    def MSB(x: Scalar[NumberT]) -> QuaLibFunctionOutput[int]:
        r"""
        Finds the index of the most significant bit in the parameter x.
        Notes:

        - Result is independent of sign, for example, +3 and -3 will return the same msb
        - The returned value will be the closet log2, rounded down.

          This is given by $\mathrm{floor}(\mathrm{log}_2(|x|))$.

          For example:

          - msb(0.1) will return -4.
          - msb(5) will return 2.
        - For an integer, msb(0) will return 0.
        - For a fixed point number, msb(0) will return -28.

        Args:
            x: a QUA fixed or a QUA int

        Returns:
            a QUA int
        """
        return _create_output_expression(Math.MSB, int, x)

    @staticmethod
    def msb(x: Scalar[NumberT]) -> QuaLibFunctionOutput[int]:
        return Math.MSB(x)

    @staticmethod
    def elu(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes the Exponential Linear Unit activation function of x
          $\mathrm{ELU(x)} = \mathrm{max}(0, x) + \mathrm{min}(0, \mathrm{exp}(x)-1)$.

        Args:
            x: a QUA fixed

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.elu, float, x)

    @staticmethod
    def aelu(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes faster an approximated Exponential Linear Unit activation function of x
          $\mathrm{aELU}(x) \sim \mathrm{ELU}(x)$

        Args:
            x: a QUA fixed

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.aelu, float, x)

    @staticmethod
    def selu(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes the Scaled Exponential Linear Unit activation function of x
        $\mathrm{SELU}(x) = s*(\mathrm{max}(0, x)+a*\mathrm{min}(0, \mathrm{exp}(x)-1))$,
        $a=1.67326324$, $s=1.05070098$

        Args:
            x: a QUA fixed

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.selu, float, x)

    @staticmethod
    def relu(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes the Rectified Linear Unit activation function of x
          $\mathrm{ReLU}(x) = \mathrm{max}(0, x)$

        Args:
            x: a QUA fixed

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.relu, float, x)

    @staticmethod
    def plrelu(x: Scalar[float], a: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes the Parametric Leaky Rectified Linear Unit activation function of x
          $\mathrm{PLReLU}(x, a) = \mathrm{max}(0, x)+a*\mathrm{min}(0, x)$

        Args:
            x: a QUA fixed
            a: a QUA fixed

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.plrelu, float, x, a)

    @staticmethod
    def lrelu(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes the Leaky Rectified Linear Unit activation function of x
          $\mathrm{LReLU}(x)=\mathrm{max}(0, x)+0.01*\mathrm{min}(0, x)$

        Args:
            x: a QUA fixed

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.lrelu, float, x)

    @staticmethod
    def sin2pi(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes $\mathrm{sin}(2 \pi x)$.
        This is more efficient than `Math.sin(2*np.pi*x)`.
        In addition, this function is immune to overflows: An overflow means that the argument gets a $\pm 16$, which does not change the result due to the periodicity of the sine function.

        Args:
            x (QUA variable of type fixed): the angle in radians

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.sin2pi, float, x)

    @staticmethod
    def cos2pi(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes $\mathrm{cos}(2 \pi x)$.
        This is more efficient than Math.cos($2 \pi x$).
        In addition, this function is immune to overflows: An overflow means that the argument gets a :math:`\pm 16`, which does not change the result due to the periodicity of the cosine function.

        Args:
            x (QUA variable of type fixed): the angle in radians

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.cos2pi, float, x)

    @staticmethod
    def atan_2pi(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes $\mathrm{1/2 \pi * atan}(x)$

        -- Available from QOP 2.4 --

        Args:
            x (a QUA fixed): The tangent ratio (opposite/adjacent)

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.atan_2pi, float, x)

    @staticmethod
    def atan2_2pi(y: Scalar[float], x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes $\mathrm{1/2 \pi * atan2}(y,x)$

        -- Available from QOP 2.4 --

        Args:
            y (a QUA fixed): The coordinate y (opposite)
            x (a QUA fixed): The coordinate x (adjacent)

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.atan2_2pi, float, y, x)

    @staticmethod
    def abs(x: ScalarOfAnyType) -> QuaLibFunctionOutput[float]:
        r"""Computes the absolute value of x

        Args:
            x: a QUA variable

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.abs, float, x)

    @staticmethod
    def sin(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes $\mathrm{sin}(x)$

        Args:
            x (QUA variable of type fixed): the angle in radians

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.sin, float, x)

    @staticmethod
    def atan(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes $\mathrm{atan}(x)$

        -- Available from QOP 2.4 --

        Args:
            x (a QUA fixed which is -1 <= x <= 1): the tangent ratio (opposite/adjacent)

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.atan, float, x)

    @staticmethod
    def atan2(y: Scalar[float], x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes $\mathrm{atan2}(y,x)$

        -- Available from QOP 2.4 --

        Args:
            y (a QUA fixed): the coordinate y (opposite)
            x (a QUA fixed): the coordinate x (adjacent)

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.atan2, float, y, x)

    @staticmethod
    def cos(x: Scalar[float]) -> QuaLibFunctionOutput[float]:
        r"""Computes $\mathrm{cos}(x)$

        Args:
            x (QUA variable of type fixed): the angle in radians

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Math.cos, float, x)

    @staticmethod
    def sum(x: Vector[NumberT]) -> QuaLibFunctionOutput[NumberT]:
        r"""Computes the sum of an array x

        Args:
            x: a QUA array

        Returns:
            the sum of the array, has same type as x
        """
        standard_x = _create_qua_vector_expression(x)
        return _create_vectors_output_expression(Math.sum, standard_x.dtype, standard_x)

    @staticmethod
    def max(x: Vector[NumberT]) -> QuaLibFunctionOutput[NumberT]:
        r"""Computes the max of an array x

        Args:
            x: a QUA array

        Returns:
            the max value of the array, has same type as x
        """
        standard_x = _create_qua_vector_expression(x)
        return _create_vectors_output_expression(Math.max, standard_x.dtype, standard_x)

    @staticmethod
    def min(x: Vector[NumberT]) -> QuaLibFunctionOutput[NumberT]:
        r"""Computes the min of an array x

        Args:
            x: a QUA array

        Returns:
            the min value of the array, has same type as x
        """
        standard_x = _create_qua_vector_expression(x)
        return _create_vectors_output_expression(Math.min, standard_x.dtype, standard_x)

    @staticmethod
    def argmax(x: Vector[NumberT]) -> QuaLibFunctionOutput[int]:
        r"""Return the index of the maximum of an array

        Args:
            x: a QUA array

        Returns:
            the index of maximum value of array, a QUA Integer
        """
        standard_x = _create_qua_vector_expression(x)
        return _create_vectors_output_expression(Math.argmax, int, standard_x)

    @staticmethod
    def argmin(x: Vector[NumberT]) -> QuaLibFunctionOutput[int]:
        r"""Return the index of the minimum of an array

        Args:
            x: a QUA array

        Returns:
            the index of minimum value of array, a QUA Integer
        """
        standard_x = _create_qua_vector_expression(x)
        return _create_vectors_output_expression(Math.argmin, int, standard_x)

    @staticmethod
    @overload
    def dot(x: Union[Vector[int], Vector[bool]], y: Union[Vector[int], Vector[bool]]) -> QuaLibFunctionOutput[int]:
        ...

    @staticmethod
    @overload
    def dot(x: Vector[float], y: VectorOfAnyType) -> QuaLibFunctionOutput[float]:
        ...

    @staticmethod
    @overload
    def dot(x: VectorOfAnyType, y: Vector[float]) -> QuaLibFunctionOutput[float]:
        ...

    @staticmethod
    def dot(x: VectorOfAnyType, y: VectorOfAnyType) -> Union[QuaLibFunctionOutput[float], QuaLibFunctionOutput[int]]:
        r"""Calculates a dot product of two QUA arrays of identical size.

        Args:
            x: a QUA array
            y: a QUA array

        Returns:
            The dot product of x and y.
            If either x or y is an array of type `Fixed`, then the result is `Fixed`. Otherwise, it is an `Int`

        Example:
            ```python
            assign(c, dot(a, b))
            ```
        """
        standard_x = _create_qua_vector_expression(x)  # type: ignore[misc]
        standard_y = _create_qua_vector_expression(y)  # type: ignore[misc]

        if issubclass(standard_x.dtype, float) or issubclass(standard_y.dtype, float):
            return _create_vectors_output_expression(Math.dot, fixed, standard_x, standard_y)
        else:
            return _create_vectors_output_expression(Math.dot, int, standard_x, standard_y)


class Cast:
    @staticmethod
    def mul_int_by_fixed(x: Scalar[int], y: Scalar[float]) -> QuaLibFunctionOutput[int]:
        r"""Multiplies an int x by a fixed y, returning an int

        Args:
            x: a QUA integer
            y: a QUA fixed

        Returns:
            a QUA int which equals x*y
        """
        return _create_output_expression(Cast.mul_int_by_fixed, int, x, y)

    @staticmethod
    def mul_fixed_by_int(x: Scalar[float], y: Scalar[int]) -> QuaLibFunctionOutput[float]:
        r"""Multiplies a fixed x by an int y, returning a fixed

        Args:
            x: a QUA fixed
            y: a QUA int

        Returns:
            a QUA fixed which equals x*y
        """
        return _create_output_expression(Cast.mul_fixed_by_int, float, x, y)

    @staticmethod
    def to_int(x: ScalarOfAnyType) -> QuaLibFunctionOutput[int]:
        r"""Casts a variable to int. Supports int, fixed or bool

        Args:
            x: a QUA variable

        Returns:
            a QUA int
        """
        return _create_output_expression(Cast.to_int, int, x)

    @staticmethod
    def to_fixed(x: ScalarOfAnyType) -> QuaLibFunctionOutput[float]:
        r"""Casts a variable to fixed. Supports int, fixed or bool

        Args:
            x: a QUA variable

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Cast.to_fixed, float, x)

    @staticmethod
    def to_bool(x: ScalarOfAnyType) -> QuaLibFunctionOutput[bool]:
        r"""Casts a variable to bool. Supports int, fixed or bool

        Args:
            x: a QUA variable

        Returns:
            a QUA bool
        """
        return _create_output_expression(Cast.to_bool, bool, x)

    @staticmethod
    def unsafe_cast_int(x: ScalarOfAnyType) -> QuaLibFunctionOutput[int]:
        r"""Treats the given input variable, bitwise, as an integer.
        For a given fixed point number, this is equivalent to multiplying by
        $2^{28}$

        Supports int, fixed or bool.

        Args:
            x: a QUA variable

        Returns:
            a QUA int
        """
        return _create_output_expression(Cast.unsafe_cast_int, int, x)

    @staticmethod
    def unsafe_cast_fixed(x: ScalarOfAnyType) -> QuaLibFunctionOutput[float]:
        r"""Treats the given input variable, bitwise, as a fixed point number.
        For a given integer, this is equivalent to multiplying by $2^{-28}$

        Supports int, fixed or bool.

        Args:
            x: a QUA variable

        Returns:
            a QUA fixed
        """
        return _create_output_expression(Cast.unsafe_cast_fixed, float, x)

    @staticmethod
    def unsafe_cast_bool(x: ScalarOfAnyType) -> QuaLibFunctionOutput[bool]:
        r"""Treats the given input variable, bitwise, as a boolean.
        A boolean is determined by the right-most bit, so for a given integer, this is
        equivalent to a parity check.

        Supports int, fixed or bool.

        Warning:
            Saving a boolean number which was unsafely cast from an integer/fixed will give the wrong value in python.

        Args:
            x: a QUA variable

        Returns:
            a QUA bool
        """
        return _create_output_expression(Cast.unsafe_cast_bool, bool, x)


class Util:
    @staticmethod
    def cond(
        condition: Scalar[bool],
        true_result: Scalar[NumberT],
        false_result: Scalar[NumberT],
    ) -> QuaLibFunctionOutput[NumberT]:
        r"""Quick conditional operation. This is equivalent to a ternary operator available in some languages:
        i.e. `a ? b : c`, meaning `b` if `a` is true, or `c` if `a` is false.
        There is less computation overhead (less latency) when running this operation relative to the if conditional.

        Example:
            ```python
            assign(var, cond(a, b, c)) #where a is a boolean expression
            ```
        """
        true_result = create_qua_scalar_expression(true_result)
        return _create_output_expression(Util.cond, true_result.dtype, condition, true_result, false_result)


class Random:
    def __init__(self, seed: Optional[Scalar[int]] = None) -> None:
        r"""A class for generating pseudo-random numbers in QUA

        Args:
            seed: Optional. An integer / QUA integer seed for the pseudo-random number
                generator.
        """
        if seed is None:
            seed = random.randrange((1 << 28) - 1)

        if isinstance(seed, int):
            self._seed = declare(int, value=seed)
        elif isinstance(seed, QuaVariable):
            self._seed = seed
        else:
            # If seed is neither an int nor a QuaVariable, we still need it as a QuaVariable (and therefore use
            # 'declare') to support assignment (used in set_seed)
            self._seed = declare(int)
            self.set_seed(seed)

    def set_seed(self, exp: Scalar[int]) -> None:
        r"""Set the seed for the pseudo-random number generator

        Args:
            exp: a QUA expression
        """
        assign(self._seed, exp)

    def rand_int(self, max_int: Scalar[int]) -> QuaLibFunctionOutput[int]:
        r"""Returns a pseudorandom integer in range [0, max_int)

        Args:
            max_int: maximum value

        :Example:
            >>> a = Random()
            >>> assign(b, a.rand_int(max_int))
        """
        return _create_output_expression(Random.rand_int, int, self._seed, max_int)

    def rand_fixed(self) -> QuaLibFunctionOutput[float]:
        r"""Returns a pseudorandom fixed in range [0.0, 1.0)

        :Example:
            >>> a = Random()
            >>> assign(b, a.rand_fixed())
        """
        return _create_output_expression(Random.rand_fixed, float, self._seed)



import warnings
import dataclasses
from typing import Set, Optional, FrozenSet, Collection

from qm.api.models.info import QuaMachineInfo
from qm.exceptions import UnsupportedCapabilitiesError


@dataclasses.dataclass(frozen=True)
class Capability:
    qop_name: str
    from_qop_version: Optional[str] = None
    name_in_exception: Optional[str] = None

    @property
    def unsupported_exception_message(self) -> str:
        if self.name_in_exception:
            capability_name = self.name_in_exception
        else:
            capability_name = self.qop_name.replace("qm.", "")
            capability_name = capability_name.replace("_", " ")

        if self.from_qop_version is None:
            return f"{capability_name} is not supported in the installed QOP version."
        else:
            return f"{capability_name} is supported from QOP {self.from_qop_version} and above."


class QopCaps:
    job_streaming_state = Capability("qm.job_streaming_state")
    multiple_inputs_for_element = Capability("qm.multiple_inputs_for_element")
    analog_delay = Capability("qm.analog_delay")
    shared_oscillators = Capability("qm.shared_oscillators")
    crosstalk = Capability("qm.crosstalk")
    shared_ports = Capability("qm.shared_ports")
    input_stream = Capability("qm.input_stream")
    new_grpc_structure = Capability("qm.new_grpc_structure")
    double_frequency = Capability("qm.double_frequency")
    command_timestamps = Capability("qm.play_tag", "2.2", "timestamping commands")
    inverted_digital_output = Capability("qm.inverted_digital_output")
    sticky_elements = Capability("qm.sticky_elements")
    octave_reset = Capability("qm.octave_reset")
    fast_frame_rotation = Capability("qm.fast_frame_rotation", "2.2")
    keeping_dc_offsets = Capability("qm.keep_dc_offsets_when_closing")
    octave_management = Capability("support_octave_mgmnt", "2.5")

    # QOP3
    qop3 = Capability("__qop3", "3.0")
    opx1000_fems_return_1_based = Capability("1_based_fem", "3.0")
    waveform_report_endpoint = Capability("qm.waveform_report_endpoint", "3.3")
    exponential_iir_filter = Capability("qm.exponential_iir_filter", "3.3")
    broadcast = Capability("qm.broadcast", "3.3")
    chunk_streaming = Capability("qm.chunk_streaming", "3.3")
    fast_frame_rotation_deprecated = Capability("qm.fast_frame_rotation_deprecated", "3.3")
    config_v2 = Capability("qm.config_v2", "3.5")
    waveform_array = Capability("qm.waveform_array", "3.5")
    exponential_dc_gain_filter = Capability("qm.exponential_dc_gain_filter", "3.5")
    multiple_streams_fetching = Capability("qm.multiple_streams_fetching", "3.5")
    external_stream = Capability("qm.external_stream", "3.5", name_in_exception="declaring an external stream")
    device_temperatures = Capability("qm.device_temperatures", "3.6")

    @staticmethod
    def get_all() -> Set[Capability]:
        # Some built in methods are also in the class dictionary, so the 'if isinstance' filters them out
        return set(cap for cap in QopCaps.__dict__.values() if isinstance(cap, Capability))

    @staticmethod
    def qop2_caps() -> Set[Capability]:
        return set(cap for cap in QopCaps.get_all() if not cap.from_qop_version or float(cap.from_qop_version) < 3)


OPX_FEM_IDX = 1


class ServerCapabilities:
    def __init__(self, supported_capabilities: Collection[Capability]) -> None:
        self._supported_capabilities = frozenset(supported_capabilities)

    # These properties exist because the previous implementation of ServerCapabilities (used until this refactor in Jan 2025)
    # included a separate property for each capability. While most usages have been updated to call the 'supports'
    # function, these specific properties are still widely used in multiple places. To avoid extensive changes, their
    # usage was retained.
    supports_double_frequency = property(lambda self: self.supports(QopCaps.double_frequency))
    supports_sticky_elements = property(lambda self: self.supports(QopCaps.sticky_elements))

    @property
    def supported_capabilities(self) -> FrozenSet[Capability]:
        return self._supported_capabilities

    # This property is defined as a function to explicitly specify a return type, ensuring compatibility with mypy.
    @property
    def fem_number_in_simulator(self) -> int:
        return OPX_FEM_IDX if self.supports(QopCaps.qop3) else 0

    def supports(self, capability: Capability) -> bool:
        return capability in self._supported_capabilities

    def validate(self, capabilities: Collection[Capability]) -> None:
        """
        Validates if the capabilities passed are supported by the server.
        Raises an UnsupportedCapabilityError for the capabilities which are not supported.
        """
        if self.supports(QopCaps.fast_frame_rotation_deprecated) and QopCaps.fast_frame_rotation in capabilities:
            warnings.warn(
                "The fast_frame_rotation is deprecated as it is no longer faster than frame_rotation_2pi "
                "(and in fact, it is less efficient). It will be removed in future versions.",
                DeprecationWarning,
            )

        unsupported_capabilities = set(capabilities) - self._supported_capabilities
        if unsupported_capabilities:
            exception_message = "\nAlso: ".join(
                [capability.unsupported_exception_message for capability in unsupported_capabilities]
            )
            raise UnsupportedCapabilitiesError(exception_message)

    @classmethod
    def build(cls, qua_implementation: Optional[QuaMachineInfo] = None) -> "ServerCapabilities":
        qop_caps = qua_implementation.capabilities if qua_implementation is not None else list()
        supported_capabilities = set(cap for cap in QopCaps.get_all() if cap.qop_name in qop_caps)

        if QopCaps.qop3.qop_name in qop_caps:
            # Doing this in case all the QOP2 caps are not added by default when it is a QOP3 machine
            supported_capabilities.update(QopCaps.qop2_caps())

        return cls(supported_capabilities)



import warnings
from typing import Union, Literal, Mapping, Optional, cast

import betterproto
from betterproto.lib.std.google.protobuf import Empty

from qm.utils import deprecation_message
from qm.exceptions import ConfigValidationException
from qm.program._dict_to_pb_converter.base_converter import BaseDictToPbConverter
from qm.program._dict_to_pb_converter.converters.octave_converter import (
    dac_port_ref_to_pb,
    _get_port_reference_with_fem,
)
from qm.program._validate_config_schema import (
    validate_oscillator,
    validate_output_tof,
    validate_used_inputs,
    validate_output_smearing,
    validate_sticky_duration,
)
from qm.type_hinting.config_types import (
    Upconverter,
    StickyConfigType,
    ElementConfigType,
    MwInputConfigType,
    PortReferenceType,
    MixInputConfigType,
    MwOutputConfigType,
    HoldOffsetConfigType,
    SingleInputConfigType,
    DigitalInputConfigType,
    InputCollectionConfigType,
    MwFemAnalogInputPortConfigType,
    TimeTaggingParametersConfigType,
)
from qm.grpc.qua_config import (
    QuaConfigSticky,
    QuaConfigMixInputs,
    QuaConfigElementDec,
    QuaConfigHoldOffset,
    QuaConfigSingleInput,
    QuaConfigElementThread,
    QuaConfigPortReference,
    QuaConfigMultipleInputs,
    QuaConfigMultipleOutputs,
    QuaConfigAdcPortReference,
    QuaConfigDacPortReference,
    QuaConfigGeneralPortReference,
    QuaConfigOutputPulseParameters,
    QuaConfigSingleInputCollection,
    QuaConfigDigitalInputPortReference,
    QuaConfigDigitalOutputPortReference,
    QuaConfigMicrowaveAnalogInputPortDec,
    QuaConfigMicrowaveInputPortReference,
    QuaConfigMicrowaveOutputPortReference,
    QuaConfigOutputPulseParametersPolarity,
)

DEFAULT_DUC_IDX = 1


class StickyElementIsNotSupported(ConfigValidationException):
    pass


class ElementConverter(BaseDictToPbConverter[ElementConfigType, QuaConfigElementDec]):
    def convert(self, input_data: ElementConfigType) -> QuaConfigElementDec:
        return self.element_to_pb(input_data)

    def element_to_pb(
        self,
        data: ElementConfigType,
    ) -> QuaConfigElementDec:
        validate_oscillator(data)
        validate_output_smearing(data)
        validate_output_tof(data)
        validate_used_inputs(data)

        element = QuaConfigElementDec()

        if "time_of_flight" in data:
            element.time_of_flight = int(data["time_of_flight"])

        if "smearing" in data:
            element.smearing = int(data["smearing"])

        if "intermediate_frequency" in data:
            element.intermediate_frequency = abs(int(data["intermediate_frequency"]))
            element.intermediate_frequency_oscillator = int(data["intermediate_frequency"])
            if self._capabilities.supports_double_frequency:
                element.intermediate_frequency_double = abs(float(data["intermediate_frequency"]))
                element.intermediate_frequency_oscillator_double = float(data["intermediate_frequency"])

            element.intermediate_frequency_negative = data["intermediate_frequency"] < 0

        if "thread" in data:
            warnings.warn(
                deprecation_message("thread", "1.2.0", "2.0.0", "Use 'core' instead"),
                DeprecationWarning,
            )
            element.thread = element_thread_to_pb(data["thread"])
        if "core" in data:
            element.thread = element_thread_to_pb(data["core"])

        if "outputs" in data:
            for k, v in data["outputs"].items():
                element.outputs[k] = adc_port_ref_to_pb(*_get_port_reference_with_fem(v))
            element.multiple_outputs = QuaConfigMultipleOutputs(port_references=element.outputs)

        if "digitalInputs" in data:
            for digital_input_k, digital_input_v in data["digitalInputs"].items():
                element.digital_inputs[digital_input_k] = digital_input_port_ref_to_pb(digital_input_v)

        if "digitalOutputs" in data:
            for digital_output_k, digital_output_v in data["digitalOutputs"].items():
                element.digital_outputs[digital_output_k] = digital_output_port_ref_to_pb(digital_output_v)

        if "operations" in data:
            for op_name, op_value in data["operations"].items():
                element.operations[op_name] = op_value

        if "singleInput" in data:
            port_ref = _get_port_reference_with_fem(data["singleInput"]["port"])
            element.single_input = single_input_to_pb(*port_ref)

        if "mixInputs" in data:
            mix_inputs = data["mixInputs"]
            element.mix_inputs = QuaConfigMixInputs(
                i=dac_port_ref_to_pb(*_get_port_reference_with_fem(mix_inputs["I"])),
                q=dac_port_ref_to_pb(*_get_port_reference_with_fem(mix_inputs["Q"])),
                mixer=mix_inputs.get("mixer", ""),
            )

            lo_frequency = mix_inputs.get("lo_frequency", 0)
            element.mix_inputs.lo_frequency = int(lo_frequency)
            if self._capabilities.supports_double_frequency:
                element.mix_inputs.lo_frequency_double = float(lo_frequency)

        if "singleInputCollection" in data:
            element.single_input_collection = QuaConfigSingleInputCollection(
                inputs={
                    k: dac_port_ref_to_pb(*_get_port_reference_with_fem(v))
                    for k, v in data["singleInputCollection"]["inputs"].items()
                }
            )

        if "multipleInputs" in data:
            element.multiple_inputs = QuaConfigMultipleInputs(
                inputs={
                    k: dac_port_ref_to_pb(*_get_port_reference_with_fem(v))
                    for k, v in data["multipleInputs"]["inputs"].items()
                }
            )

        if "MWInput" in data:
            element.microwave_input = QuaConfigMicrowaveInputPortReference(
                port=dac_port_ref_to_pb(*_get_port_reference_with_fem(data["MWInput"]["port"])),
                upconverter=data["MWInput"].get("upconverter", DEFAULT_DUC_IDX),
            )
        if "MWOutput" in data:
            mw_output = data["MWOutput"]
            element.microwave_output = QuaConfigMicrowaveOutputPortReference(
                port=adc_port_ref_to_pb(*_get_port_reference_with_fem(mw_output["port"])),
            )

        if "oscillator" in data:
            element.named_oscillator = data["oscillator"]
        elif "intermediate_frequency" not in data:
            element.no_oscillator = Empty()

        if "sticky" in data:
            if "duration" in data["sticky"]:
                validate_sticky_duration(data["sticky"]["duration"])
            if self._capabilities.supports_sticky_elements:
                element.sticky = QuaConfigSticky(
                    analog=data["sticky"].get("analog", True),
                    digital=data["sticky"].get("digital", False),
                    duration=int(data["sticky"].get("duration", 4) / 4),
                )
            else:
                if "digital" in data["sticky"] and data["sticky"]["digital"]:
                    raise StickyElementIsNotSupported("Not supported")
                element.hold_offset = QuaConfigHoldOffset(duration=int(data["sticky"].get("duration", 4) / 4))

        elif "hold_offset" in data:
            if self._capabilities.supports_sticky_elements:
                element.sticky = QuaConfigSticky(
                    analog=True,
                    digital=False,
                    duration=data["hold_offset"].get("duration", 1),
                )
            else:
                element.hold_offset = QuaConfigHoldOffset(duration=data["hold_offset"]["duration"])

        if "outputPulseParameters" in data:
            warnings.warn(
                deprecation_message("outputPulseParameters", "1.2.0", "2.0.0", "Use timeTaggingParameters instead"),
                DeprecationWarning,
            )
            element.output_pulse_parameters = self.create_time_tagging_parameters(data["outputPulseParameters"])
        if "timeTaggingParameters" in data:
            element.output_pulse_parameters = self.create_time_tagging_parameters(data["timeTaggingParameters"])

        rf_inputs = data.get("RF_inputs", {})
        for k, (device, port) in rf_inputs.items():
            element.rf_inputs[k] = QuaConfigGeneralPortReference(device_name=device, port=port)

        rf_outputs = data.get("RF_outputs", {})
        for k, (device, port) in rf_outputs.items():
            element.rf_outputs[k] = QuaConfigGeneralPortReference(device_name=device, port=port)
        return element

    @staticmethod
    def create_time_tagging_parameters(data: TimeTaggingParametersConfigType) -> QuaConfigOutputPulseParameters:
        return QuaConfigOutputPulseParameters(
            signal_threshold=data["signalThreshold"],
            signal_polarity=ElementConverter._create_signal_polarity(data["signalPolarity"]),
            derivative_threshold=data["derivativeThreshold"],
            derivative_polarity=ElementConverter._create_signal_polarity(data["derivativePolarity"]),
        )

    @staticmethod
    def _create_signal_polarity(polarity: str) -> QuaConfigOutputPulseParametersPolarity:
        polarity = polarity.upper()
        if polarity in {"ABOVE", "ASCENDING"}:
            if polarity == "ASCENDING":
                warnings.warn(
                    deprecation_message("ASCENDING", "1.2.2", "2.0.0", "Use 'ABOVE' instead"), DeprecationWarning
                )
            return QuaConfigOutputPulseParametersPolarity.ASCENDING  # type: ignore[return-value]
        elif polarity in {"BELOW", "DESCENDING"}:
            if polarity == "DESCENDING":
                warnings.warn(
                    deprecation_message("DESCENDING", "1.2.2", "2.0.0", "Use 'BELOW' instead"), DeprecationWarning
                )
            return QuaConfigOutputPulseParametersPolarity.DESCENDING  # type: ignore[return-value]
        else:
            raise ConfigValidationException(f"Invalid signal polarity: {polarity}")

    def deconvert(self, output_data: QuaConfigElementDec) -> ElementConfigType:
        element_output = betterproto.which_one_of(output_data, "element_outputs_one_of")[1]
        assert (
            isinstance(element_output, (QuaConfigMultipleOutputs, QuaConfigMicrowaveOutputPortReference))
            or element_output is None
        )
        element_config_data: ElementConfigType = {
            "digitalInputs": _deconvert_inputs(output_data.digital_inputs),
            "digitalOutputs": _deconvert_digital_output(output_data.digital_outputs),
            "outputs": _deconvert_element_output(output_data.outputs, element_output),
            "operations": output_data.operations,
            "hold_offset": _deconvert_hold_offset(output_data.hold_offset),
            "sticky": _deconvert_sticky(output_data.sticky),
        }
        if betterproto.serialized_on_wire(output_data.thread):
            element_config_data["core"] = _deconvert_element_thread(output_data.thread)
        if betterproto.serialized_on_wire(output_data.output_pulse_parameters):
            element_config_data["timeTaggingParameters"] = _deconvert_time_tagging_params(
                output_data.output_pulse_parameters
            )
        input_value = betterproto.which_one_of(output_data, "element_inputs_one_of")[1]
        if isinstance(input_value, QuaConfigSingleInput):
            element_config_data["singleInput"] = _deconvert_single_inputs(input_value)
        elif isinstance(input_value, QuaConfigMixInputs):
            element_config_data["mixInputs"] = _deconvert_mix_inputs(input_value)
        elif isinstance(input_value, QuaConfigSingleInputCollection):
            element_config_data["singleInputCollection"] = _deconvert_single_input_collection(input_value)
        elif isinstance(input_value, QuaConfigMultipleInputs):
            element_config_data["multipleInputs"] = _deconvert_multiple_inputs(input_value)
        elif isinstance(input_value, QuaConfigMicrowaveInputPortReference):
            element_config_data["MWInput"] = _deconvert_element_mw_input(input_value)

        output_value = betterproto.which_one_of(output_data, "element_outputs_one_of")[1]
        if isinstance(output_value, QuaConfigMicrowaveOutputPortReference):
            element_config_data["MWOutput"] = _deconvert_element_mw_output(output_value)

        if output_data.smearing is not None:
            element_config_data["smearing"] = output_data.smearing
        if output_data.time_of_flight is not None:
            element_config_data["time_of_flight"] = output_data.time_of_flight
        if output_data.measurement_qe:
            element_config_data["measurement_qe"] = output_data.measurement_qe

        oscillator_key, oscillator = betterproto.which_one_of(output_data, "oscillator_one_of")
        if oscillator_key == "named_oscillator":
            assert isinstance(oscillator, str)
            element_config_data["oscillator"] = oscillator
        else:
            sign = (-1) ** output_data.intermediate_frequency_negative
            if output_data.intermediate_frequency_double:
                element_config_data["intermediate_frequency"] = abs(output_data.intermediate_frequency_double) * sign
            elif output_data.intermediate_frequency is not None:
                element_config_data["intermediate_frequency"] = abs(output_data.intermediate_frequency) * sign

        return element_config_data


def element_thread_to_pb(name: str) -> QuaConfigElementThread:
    return QuaConfigElementThread(thread_name=name)


def single_input_to_pb(controller: str, fem: int, number: int) -> QuaConfigSingleInput:
    return QuaConfigSingleInput(port=dac_port_ref_to_pb(controller, fem, number))


def adc_port_ref_to_pb(controller: str, fem: int, number: int) -> QuaConfigAdcPortReference:
    return QuaConfigAdcPortReference(controller=controller, fem=fem, number=number)


def port_ref_to_pb(controller: str, fem: int, number: int) -> QuaConfigPortReference:
    return QuaConfigPortReference(controller=controller, fem=fem, number=number)


def digital_input_port_ref_to_pb(data: DigitalInputConfigType) -> QuaConfigDigitalInputPortReference:
    digital_input = QuaConfigDigitalInputPortReference(
        delay=int(data["delay"]),
        buffer=int(data["buffer"]),
    )
    if "port" in data:
        digital_input.port = port_ref_to_pb(*_get_port_reference_with_fem(data["port"]))

    return digital_input


def digital_output_port_ref_to_pb(data: PortReferenceType) -> QuaConfigDigitalOutputPortReference:
    return QuaConfigDigitalOutputPortReference(port=port_ref_to_pb(*_get_port_reference_with_fem(data)))


def _deconvert_polarity(polarity: QuaConfigOutputPulseParametersPolarity) -> Literal["ABOVE", "BELOW"]:
    if polarity == QuaConfigOutputPulseParametersPolarity.ASCENDING:
        return "ABOVE"
    if polarity == QuaConfigOutputPulseParametersPolarity.DESCENDING:
        return "BELOW"
    raise ValueError(f"Unknown polarity - {polarity}")


def _deconvert_time_tagging_params(data: QuaConfigOutputPulseParameters) -> TimeTaggingParametersConfigType:
    to_return: TimeTaggingParametersConfigType = {
        "signalThreshold": data.signal_threshold,
        "signalPolarity": _deconvert_polarity(data.signal_polarity),
        "derivativeThreshold": data.derivative_threshold,
        "derivativePolarity": _deconvert_polarity(data.derivative_polarity),
    }
    return to_return


def _deconvert_mix_inputs(mix_inputs: QuaConfigMixInputs) -> MixInputConfigType:
    res: MixInputConfigType = {
        "I": _deconvert_port_reference(mix_inputs.i),
        "Q": _deconvert_port_reference(mix_inputs.q),
    }

    mixer = mix_inputs.mixer
    if mixer is not None:
        res["mixer"] = mixer

    if mix_inputs.lo_frequency_double:
        res["lo_frequency"] = mix_inputs.lo_frequency_double
    else:
        res["lo_frequency"] = float(mix_inputs.lo_frequency)

    return res


def _deconvert_single_inputs(single: QuaConfigSingleInput) -> SingleInputConfigType:
    return {"port": _deconvert_port_reference(single.port)}


def _deconvert_hold_offset(hold_offset: QuaConfigHoldOffset) -> HoldOffsetConfigType:
    return {"duration": hold_offset.duration}


def _deconvert_sticky(sticky: QuaConfigSticky) -> StickyConfigType:
    res: StickyConfigType = {
        "analog": sticky.analog,
        "digital": sticky.digital,
        "duration": max(sticky.duration, 1) * 4,
    }
    return res


def _deconvert_element_mw_input(data: QuaConfigMicrowaveInputPortReference) -> MwInputConfigType:
    return {
        "port": _deconvert_port_reference(data.port),
        "upconverter": cast(Upconverter, data.upconverter),
    }


def _deconvert_element_mw_output(data: QuaConfigMicrowaveOutputPortReference) -> MwOutputConfigType:
    return {
        "port": _deconvert_port_reference(data.port),
    }


def _deconvert_element_thread(element_thread: QuaConfigElementThread) -> str:
    return element_thread.thread_name


def _deconvert_port_reference(
    data: Union[QuaConfigAdcPortReference, QuaConfigDacPortReference, QuaConfigPortReference],
) -> PortReferenceType:
    if data.fem:
        return data.controller, data.fem, data.number
    else:
        return data.controller, data.number


def _deconvert_mw_analog_inputs(
    inputs: dict[int, QuaConfigMicrowaveAnalogInputPortDec],
) -> Mapping[Union[int, str], MwFemAnalogInputPortConfigType]:
    return {idx: _deconvert_single_mw_analog_input(_input) for idx, _input in inputs.items()}


def _deconvert_inputs(inputs: dict[str, QuaConfigDigitalInputPortReference]) -> dict[str, DigitalInputConfigType]:
    ret: dict[str, DigitalInputConfigType] = {}
    for name, data in inputs.items():
        ret[name] = {"delay": data.delay, "buffer": data.buffer, "port": _deconvert_port_reference(data.port)}
    return ret


def _deconvert_digital_output(outputs: dict[str, QuaConfigDigitalOutputPortReference]) -> dict[str, PortReferenceType]:
    ret = {}
    for name, data in outputs.items():
        ret[name] = _deconvert_port_reference(data.port)

    return ret


def _deconvert_single_input_collection(data: QuaConfigSingleInputCollection) -> InputCollectionConfigType:
    temp = {}
    for name, input_info in data.inputs.items():
        temp[name] = _deconvert_port_reference(input_info)

    res: InputCollectionConfigType = {"inputs": temp}
    return res


def _deconvert_multiple_inputs(data: QuaConfigMultipleInputs) -> InputCollectionConfigType:
    temp = {}
    for name, input_info in data.inputs.items():
        temp[name] = _deconvert_port_reference(input_info)

    res: InputCollectionConfigType = {"inputs": temp}
    return res


def _deconvert_single_mw_analog_input(data: QuaConfigMicrowaveAnalogInputPortDec) -> MwFemAnalogInputPortConfigType:
    return cast(
        MwFemAnalogInputPortConfigType,
        {
            "band": cast(Literal[1, 3, 3], data.band),
            "shareable": data.shareable,
            "gain_db": data.gain_db,
            "sampling_rate": data.sampling_rate,
            "downconverter_frequency": data.downconverter.frequency,
        },
    )


def _deconvert_element_output(
    outputs: dict[str, QuaConfigAdcPortReference],
    multiple_outputs: Optional[Union[QuaConfigMultipleOutputs, QuaConfigMicrowaveOutputPortReference]],
) -> dict[str, PortReferenceType]:
    if isinstance(multiple_outputs, QuaConfigMultipleOutputs):
        return {name: _deconvert_port_reference(data) for name, data in multiple_outputs.port_references.items()}
    return {name: _deconvert_port_reference(data) for name, data in outputs.items()}



import warnings

from qm._loc import _get_loc
from qm.utils import deprecation_message
from qm.qua._scope_management.scopes_manager import scopes_manager
from qm.grpc.qua import (
    QuaProgramAnyStatement,
    QuaProgramResetPhaseStatement,
    QuaProgramQuantumElementReference,
    QuaProgramResetGlobalPhaseStatement,
)


def reset_phase(element: str) -> None:
    warnings.warn(
        deprecation_message(
            method="reset_phase",
            deprecated_in="1.2.2",
            removed_in="2.0.0",
            details="reset_if_phase instead.",
        ),
        DeprecationWarning,
        stacklevel=2,
    )
    reset_if_phase(element)


def reset_if_phase(element: str) -> None:
    r"""
    Resets the intermediate frequency phase of the oscillator associated with `element`,
    setting the phase of the next pulse to absolute zero.
    This sets the phase of the currently playing intermediate frequency
    to the value it had at the beginning of the program (t=0).

    Note:

        * The phase will only be set to zero when the next play or align command is executed on the element.
        * Reset phase will only reset the phase of the intermediate frequency (:math:`\\omega_{IF}`) currently in use.

    Args:
        element: an element
    """

    loc = _get_loc()
    statement = QuaProgramResetPhaseStatement(loc=loc, qe=QuaProgramQuantumElementReference(name=element, loc=loc))
    scopes_manager.append_statement(QuaProgramAnyStatement(reset_phase=statement))


def reset_global_phase() -> None:
    """
    Resets the global phase of all the elements in the program.
    This will reset both the intermediate frequency phase and the upconverters/downconverters in use.
    """
    statement = QuaProgramResetGlobalPhaseStatement(loc=_get_loc())
    scopes_manager.append_statement(QuaProgramAnyStatement(reset_global_phase=statement))



from qm.qua._dsl.other import L
from qm.qua._expressions import fixed
from qm.exceptions import QmQuaException
from qm.qua._expressions import IO1, IO2
from qm.qua._dsl.broadcast import broadcast
from qm.qua._qua_struct import QuaArray, qua_struct
from qm.qua._dsl.amplitude import AmpValuesType, amp
from qm.qua._expressions import QuaVariable as Variable
from qm.qua._dsl.pulses_utils import ramp_to_zero, load_waveform
from qm.qua._dsl.measure.measure import MeasurePulseType, measure
from qm.qua._dsl.play import ChirpType, PlayPulseType, play, ramp
from qm.qua._dsl.wait import wait, align, pause, wait_for_trigger
from qm.qua._dsl.stream_processing.stream_processing_utils import bins
from qm.qua._dsl.measure.analog_measure_process import AnalogMeasureProcess
from qm.qua._dsl.measure.digital_measure_process import DigitalMeasureProcess
from qm.qua._dsl.stream_processing.map_functions.map_functions import FUNCTIONS
from qm.qua._dsl.phase_reset import reset_phase, reset_if_phase, reset_global_phase
from qm.qua._dsl._type_hints import OneOrMore, MessageVarType, MessageExpressionType
from qm.qua._dsl.stream_processing.stream_processing import StreamType, declare_stream
from qm.qua._dsl.calibration_params_update import set_dc_offset, update_frequency, update_correction
from qm.qua.lib import Cast, Math, Util, Random, call_library_function, call_vectors_library_function
from qm.qua._dsl.frame_rotation import reset_frame, frame_rotation, frame_rotation_2pi, fast_frame_rotation
from qm.qua._dsl.measure.measure_process_factories import demod, counting, dual_demod, integration, time_tagging
from qm.qua._dsl.external_stream import (
    QuaStreamDirection,
    declare_external_stream,
    send_to_external_stream,
    receive_from_external_stream,
)
from qm.qua._dsl.variable_handling import (
    DeclarationType,
    save,
    assign,
    declare,
    declare_struct,
    advance_input_stream,
    declare_input_stream,
)
from qm.qua._dsl.scope_functions import (
    if_,
    for_,
    case_,
    elif_,
    else_,
    while_,
    program,
    switch_,
    default_,
    for_each_,
    infinite_loop_,
    port_condition,
    strict_timing_,
    stream_processing,
)

__all__ = [
    "L",
    "if_",
    "amp",
    "IO1",
    "IO2",
    "Math",
    "Cast",
    "Util",
    "play",
    "wait",
    "save",
    "for_",
    "ramp",
    "bins",
    "fixed",
    "pause",
    "align",
    "case_",
    "else_",
    "elif_",
    "demod",
    "Random",
    "assign",
    "while_",
    "program",
    "measure",
    "switch_",
    "declare",
    "default_",
    "counting",
    "Variable",
    "broadcast",
    "ChirpType",
    "for_each_",
    "FUNCTIONS",
    "StreamType",
    "dual_demod",
    "reset_frame",
    "reset_phase",
    "integration",
    "ramp_to_zero",
    "time_tagging",
    "load_waveform",
    "AmpValuesType",
    "PlayPulseType",
    "set_dc_offset",
    "reset_if_phase",
    "MessageVarType",
    "frame_rotation",
    "QmQuaException",
    "infinite_loop_",
    "port_condition",
    "declare_stream",
    "strict_timing_",
    "DeclarationType",
    "MeasurePulseType",
    "update_frequency",
    "wait_for_trigger",
    "update_correction",
    "stream_processing",
    "frame_rotation_2pi",
    "reset_global_phase",
    "fast_frame_rotation",
    "AnalogMeasureProcess",
    "declare_input_stream",
    "advance_input_stream",
    "MessageExpressionType",
    "DigitalMeasureProcess",
    "call_library_function",
    "call_vectors_library_function",
    "qua_struct",
    "QuaArray",
    "declare_struct",
    "declare_external_stream",
    "QuaStreamDirection",
    "receive_from_external_stream",
    "send_to_external_stream",
]



import numbers
from typing import List, Tuple, Union, cast

from qm.exceptions import ConfigValidationException
from qm.utils.list_compression_utils import split_list_to_chunks
from qm.type_hinting.config_types import IntegrationWeightConfigType
from qm.program._dict_to_pb_converter.base_converter import BaseDictToPbConverter
from qm.grpc.qua_config import QuaConfigIntegrationWeightDec, QuaConfigIntegrationWeightSample


class IntegrationWeightsConverter(BaseDictToPbConverter[IntegrationWeightConfigType, QuaConfigIntegrationWeightDec]):
    def convert(self, input_data: IntegrationWeightConfigType) -> QuaConfigIntegrationWeightDec:
        return self.integration_weights_to_pb(input_data)

    @staticmethod
    def integration_weights_to_pb(data: IntegrationWeightConfigType) -> QuaConfigIntegrationWeightDec:
        iw = QuaConfigIntegrationWeightDec(
            cosine=build_iw_sample(data["cosine"]),
            sine=build_iw_sample(data["sine"]),
        )
        return iw

    def deconvert(self, output_data: QuaConfigIntegrationWeightDec) -> IntegrationWeightConfigType:
        return {
            "cosine": [(s.value, s.length) for s in output_data.cosine],
            "sine": [(s.value, s.length) for s in output_data.sine],
        }


def build_iw_sample(data: Union[List[Tuple[float, int]], List[float]]) -> List[QuaConfigIntegrationWeightSample]:
    clean_data = _standardize_iw_data(data)
    return [QuaConfigIntegrationWeightSample(value=s[0], length=int(s[1])) for s in clean_data]


def _standardize_iw_data(data: Union[List[Tuple[float, int]], List[float]]) -> List[Tuple[float, int]]:
    if len(data) == 0 or isinstance(data[0], (tuple, list)):
        to_return = []
        for x in data:
            x = cast(Tuple[float, int], x)
            to_return.append((x[0], x[1]))
        return to_return

    if isinstance(data[0], numbers.Number):
        if len(data) == 2:
            d0, d1 = cast(Tuple[float, int], data)
            return [(float(d0), int(d1))]

        data = cast(List[float], data)
        chunks = split_list_to_chunks([round(2**-15 * round(s * 2**15), 20) for s in data])
        new_data: List[Tuple[float, int]] = []
        for chunk in chunks:
            if chunk.accepts_different:
                new_data.extend([(float(u), 4) for u in chunk.data])
            else:
                new_data.append((chunk.first, 4 * len(chunk)))
        return new_data

    raise ConfigValidationException(f"Invalid IW data, data must be a list of numbers or 2-tuples, got {data}.")



import logging
import warnings
from typing import List, Tuple, Union, Optional, overload

from qm._loc import _get_loc
from qm.exceptions import QmQuaException
from qm.qua._expressions import QuaVariable
from qm.qua._dsl.amplitude import AmpValuesType
from qm.qua._dsl._utils import _standardize_timestamp_label
from qm.qua._dsl.measure.measure_process_factories import demod
from qm.qua._scope_management.scopes_manager import scopes_manager
from qm.qua._dsl.measure.analog_measure_process import MeasureProcessAbstract
from qm.qua._dsl.stream_processing.stream_processing import StreamType, ResultStreamSource, declare_stream
from qm.grpc.qua import (
    QuaProgramAnyStatement,
    QuaProgramPulseReference,
    QuaProgramMeasureStatement,
    QuaProgramQuantumElementReference,
)

logger = logging.getLogger(__name__)

MeasurePulseType = Union[str, Tuple[str, AmpValuesType]]


@overload
def measure(
    pulse: MeasurePulseType,
    element: str,
    *outputs: Union[Tuple[str, QuaVariable[float]], Tuple[str, str, QuaVariable[float]], MeasureProcessAbstract],
    timestamp_stream: Optional[StreamType] = None,
    adc_stream: Optional[StreamType] = None,
) -> None:
    pass


@overload
def measure(
    pulse: MeasurePulseType,
    element: str,
    stream: Optional[StreamType],
    *outputs: Union[Tuple[str, QuaVariable[float]], Tuple[str, str, QuaVariable[float]], MeasureProcessAbstract],
    timestamp_stream: Optional[StreamType] = None,
    adc_stream: Optional[StreamType] = None,
) -> None:
    pass


def measure(
    pulse: MeasurePulseType,
    element: str,
    *outputs: Union[
        Tuple[str, QuaVariable[float]],
        Tuple[str, str, QuaVariable[float]],
        MeasureProcessAbstract,
        Optional[StreamType],
    ],
    stream: Optional[StreamType] = None,
    timestamp_stream: Optional[StreamType] = None,
    adc_stream: Optional[StreamType] = None,
) -> None:
    """Perform a measurement of `element` using `pulse` based on 'operation' as defined in the 'element'.

    An element for which a measurement is applied must have outputs defined in the configuration.

    A measurement consists of:

    1. playing an operation to the element (identical to a :func:`play` statement)

    2. waiting for a duration of time defined as the ``time_of_flight``
       in the configuration of the element, and then sampling
       the returning pulse.
       The OPX input to be sampled is defined in the configuration of the element.

    3. Processing the acquired data according to a parameter defined in the measure command,
        including Demodulation, Integration and Time Tagging.

    For a more detailed description of the measurement operation, see
    [Measure Statement Features](../../Guides/features.md#measure-statement-features)

    Args:
        pulse (str): The name of an `operation` to be performed, as defined in the element in the quantum machine configuration.
            Pulse must have a ``measurement`` operation. Can also be multiplied by an [amp][qm.qua.amp].
        element (str): name of the element, as defined in the quantum machine configuration. The element must have outputs.
        *outputs (tuple): A parameter specifying the processing to be done on the ADC data, there are multiple options available, including demod(), integration() & time_tagging().
        stream (Union[str, ResultStreamSource]): Deprecated and replaced by `adc_stream`.
        timestamp_stream (Union[str, ResultStreamSource]): (Supported from QOP 2.2) Adding a `timestamp_stream` argument will save the time at which the operation occurred to a stream.
            If the `timestamp_stream` is a string ``label``, then the timestamp handle can be retrieved with [qm.results.StreamsManager][] with the same ``label``.
        adc_stream (Union[str, ResultStreamSource]): The stream variable into which the raw ADC data will be saved.
            You can receive the results with [qm.QmJob.result_handles.get("name")][qm.jobs.running_qm_job.RunningQmJob.result_handles].
            A string name can also be used. In this case, the name of the result
            handle should be suffixed by ``_input1`` for data from analog input 1 and ``_input2`` for data from analog input 2.

            If ``adc_stream`` is set to ``None``, nothing will not be saved.
            The raw results will be saved as long as the digital pulse that is played with pulse is high.

            !!! Warning:

                Streaming adc data without declaring the stream with `declare_stream(adc_trace=true)` might cause performance issues

    Example:
        ```python
        with program() as prog:
            I = declare(fixed)
            Q = declare(fixed)
            adc_st = declare_stream(adc_trace=True)

            # measure by playing 'meas_pulse' to element 'resonator', do not save raw results.
            # demodulate data from "out1" port of 'resonator' using 'cos_weights' and store result in I, and also
            # demodulate data from "out1" port of 'resonator' using 'sin_weights' and store result in Q
            measure('meas_pulse', 'resonator', demod.full("cos_weights", I, "out1"), demod.full("sin_weights", Q, "out1"))

            # measure by playing 'meas_pulse' to element 'resonator', save raw results to `adc_st`
            # demodulate data from 'out1' port of 'resonator' using 'optimized_weights' and store result in I
            measure('meas_pulse', 'resonator', adc_st, demod.full("optimized_weights", I, "out1"))
            with stream_processing():
                adc_st.input1().save_all("raw_adc_stream")

        from qm import QuantumMachinesManager
        qm = QuantumMachinesManager().open_qm(config)
        job = qm.execute(prog)
        # ... we wait for the results to be ready...
        job.result_handles.wait_for_all_values()
        # raw results can be retrieved as follows (here job is a QmJob object:
        raw_I_handle = job.result_handles.get("raw_adc_stream")
        ```
    """
    if stream is None:
        if len(outputs) > 0:
            if isinstance(outputs[0], (ResultStreamSource, str)):
                adc_stream = outputs[0]
                if isinstance(adc_stream, ResultStreamSource):
                    warnings.warn(
                        "Saving an adc stream now requires defining it at the end of the measure command with the `adc_stream` argument, e.g. `adc_stream=adc_st`. The current syntax is deprecated and will be removed in a 1.3.",
                        DeprecationWarning,
                        stacklevel=2,
                    )
                else:
                    warnings.warn(
                        f"Saving an adc stream now requires defining it at the end of the measure command with the `adc_stream` argument, e.g. `adc_stream='{adc_stream}'`. The current syntax is deprecated and will be removed in a 1.3.",
                        DeprecationWarning,
                        stacklevel=2,
                    )
                outputs = outputs[1:]
            elif outputs[0] is None:
                warnings.warn(
                    "Putting `None` to indicate no adc streaming is no longer required, please remove it from the measure call.",
                    DeprecationWarning,
                    stacklevel=2,
                )
                outputs = outputs[1:]
    else:
        warnings.warn(
            "The `stream` argument is deprecated and will be removed in a 1.3. Use the `adc_stream` argument instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        adc_stream = stream

    measure_process: List[MeasureProcessAbstract] = []
    for i, output in enumerate(outputs):
        if isinstance(output, tuple):
            if len(output) == 2:
                iw, target = output
                measure_process.append(demod.full(iw, target, ""))
            elif len(output) == 3:
                iw, element_output, target = output
                measure_process.append(demod.full(iw, target, element_output))
            else:
                raise QmQuaException(
                    "Each output must be a tuple of (integration weight, output name, variable name), but output "
                    + str(i + 1)
                    + " is invalid"
                )
        elif isinstance(output, MeasureProcessAbstract):
            measure_process.append(output)
        else:
            raise TypeError(f"Invalid output type: {type(output)}")

    _adc_stream: Optional[ResultStreamSource] = None
    if isinstance(adc_stream, str):
        _adc_stream = _declare_legacy_adc(adc_stream)
    else:
        if adc_stream is not None and (not isinstance(adc_stream, ResultStreamSource)):
            raise QmQuaException("stream object is not of the right type")
        _adc_stream = adc_stream

    if _adc_stream and not _adc_stream.is_adc_trace:
        logger.warning(
            "Streaming adc data without declaring the stream with "
            "`declare_stream(adc_trace=true)` might cause performance issues"
        )
    timestamp_label = _standardize_timestamp_label(timestamp_stream)
    processes = [x.unwrapped for x in measure_process]

    amp = None
    if isinstance(pulse, tuple):
        pulse, amp = pulse

    loc = _get_loc()
    statement = QuaProgramAnyStatement(
        measure=QuaProgramMeasureStatement(
            loc=loc,
            pulse=QuaProgramPulseReference(name=pulse, loc=loc),
            qe=QuaProgramQuantumElementReference(name=element, loc=loc),
        )
    )
    if _adc_stream is not None:
        statement.measure.stream_as = _adc_stream.get_var_name()

    for analog_process in processes:
        statement.measure.measure_processes.append(analog_process)

    if amp is not None:
        statement.measure.amp.loc = loc
        statement.measure.amp.v0 = amp[0]
        for i in range(1, 4, 1):
            if amp[i] is not None:
                setattr(statement.measure.amp, "v" + str(i), amp[i])

    if timestamp_label is not None:
        statement.measure.timestamp_label = timestamp_label

    scopes_manager.append_statement(statement)


def _declare_legacy_adc(tag: str) -> ResultStreamSource:
    program_scope = scopes_manager.program_scope
    result_object = program_scope.declared_streams.get(tag, None)
    if result_object is None:
        result_object = declare_stream(adc_trace=True)
        program_scope.add_stream_declaration(tag, result_object)
        result_object.input1().with_timestamps()._auto_save_all(tag + "_input1")
        result_object.input2().with_timestamps()._auto_save_all(tag + "_input2")

    return result_object



import logging
from typing import Union, Optional, cast

from qm.exceptions import QmInvalidSchemaError
from qm.type_hinting.general import NumpyArray
from qm._stream_results._single_stream_fetchers._base_single_stream_fetcher import (
    VERY_LONG_TIME,
    BaseSingleStreamFetcher,
    NumpyArrayOrSingleValue,
)

logger = logging.getLogger(__name__)


class SingleStreamSingleResultFetcher(BaseSingleStreamFetcher[Optional[NumpyArrayOrSingleValue]]):
    """A handle to a result of a pipeline terminating with ``save``"""

    def _validate_schema(self) -> None:
        if not self._schema.is_single:
            raise QmInvalidSchemaError("expecting a single-result schema")

    def wait_for_values(self, count: int = 1, timeout: float = VERY_LONG_TIME) -> None:
        if count != 1:
            raise RuntimeError("single result can wait only for a single value")
        super().wait_for_values(1, timeout)

    def fetch_all(
        self, *, check_for_errors: bool = True, flat_struct: bool = False
    ) -> Optional[NumpyArrayOrSingleValue]:
        """
        Fetches the results from the current result stream saved in server memory.
        The result stream is populated by the `save()` statement.

        Args:
            check_for_errors: If true, the function would also check whether run-time errors happened during the
                program execution and would write to the logger an error message.
            flat_struct: Results will have a flat structure - dimensions will be part of the shape and not of the type

        Returns:
            The current result from the stream. This can be a single numpy scalar or a numpy array, which depends
            on the stream operators that were used. For example, using a buffer or 'with_timestamps' will return a
            numpy array.
        """
        return self.fetch(0, flat_struct=flat_struct, check_for_errors=check_for_errors)

    def fetch(
        self,
        item: Union[int, slice],
        *,
        check_for_errors: bool = True,
        flat_struct: bool = False,
        timeout: Optional[float] = None,
    ) -> Optional[NumpyArrayOrSingleValue]:
        """
        Fetches the results from the current result stream saved in server memory.
        The result stream is populated by the `save()` statement.

        Args:
            item: The index, or a slice indicating a range, of the result in the stream.
            check_for_errors: If true, the function would also check whether run-time errors happened during the
                program execution and would write to the logger an error message.
            flat_struct: Results will have a flat structure - dimensions will be part of the shape and not of the type
            timeout: Timeout for waiting in seconds.

        Returns:
            The current result. This can be a single numpy scalar or a numpy array, which depends
            on the stream operators that were used. For example, using a buffer or 'with_timestamps' will return a
            numpy array.

        Example:
            ```python
            res.fetch() # return the item in the top position
            ```
        """
        if item not in [0, slice(None), slice(1), slice(0, 1), slice(0, 1, 1)]:
            logger.warning("Fetching single result will always return the single value")
        return super().fetch(0, check_for_errors=check_for_errors, flat_struct=flat_struct, timeout=timeout)

    @staticmethod
    def _postprocess(fetched_data: NumpyArray, flat_struct: bool) -> Optional[NumpyArrayOrSingleValue]:
        """
        Peel a single layer of the fetched data array, if possible.
        It does not mean that the returned value is a scalar - it can be still an array. In the case of an array that is
        "double layered" (happens when with_timestamps is used for example), the returned value will be the underlying array.
        Also, cases where a regular array is passed with flat_struct=True are handled here, so that the returned value is the
        same as the input value.
        """
        # We assume here that we have at most a single value in array (or nothing)
        if len(fetched_data) == 0:
            logger.warning("Nothing to fetch: no results were found. Please wait until the results are ready.")
            return None
        if flat_struct:
            data = fetched_data
        else:
            data = fetched_data[0]

        if len(data) == 1:
            to_return = data[0]
        else:
            to_return = data
        return cast(NumpyArrayOrSingleValue, to_return)



from math import ceil
from typing import List, Iterable

from betterproto.lib.google.protobuf import Value, ListValue

from qm.type_hinting import Number

_ARRAY_SYMBOL = "@array"


def create_array(iterable: Iterable[Number]) -> Value:
    values = (_ARRAY_SYMBOL,) + tuple(str(item) for item in iterable)
    return Value(list_value=ListValue(values=[Value(string_value=s) for s in values]))


def bins(start: Number, end: Number, number_of_bins: float) -> List[List[Number]]:
    bin_size = ceil((end - start + 1) / float(number_of_bins))
    bins_list: List[List[Number]] = []
    while start < end:
        step_end = start + bin_size - 1
        if step_end >= end:
            step_end = end
        bins_list = bins_list + [[start, step_end]]
        start += bin_size
    return bins_list



from collections import defaultdict
from typing import Dict, Type, Union, Mapping, Optional, Sequence, AsyncIterator

import betterproto

from qm.api.v2.base_api_v2 import BaseApiV2
from qm.exceptions import DataFetchingError
from qm.api.models.server_details import ConnectionDetails
from qm.api.models.jobs import JobNamedResult, JobStreamingState, JobResultItemSchema, JobNamedResultHeader
from qm.grpc.v2 import (
    Range,
    JobServiceStub,
    JobExecutionStatus,
    GetNamedResultRequest,
    GetNamedResultResponse,
    GetNamedResultsRequest,
    GetJobResultStateRequest,
    GetJobResultSchemaRequest,
    GetNamedResultsRequestOutput,
    JobServiceGetJobStatusRequest,
    GetJobNamedResultHeaderRequest,
    GetJobNamedResultsHeadersRequest,
    GetJobNamedResultsHeadersRequestOutput,
    GetNamedResultResponseGetNamedResultResponseError,
    GetNamedResultResponseGetNamedResultResponseSuccess,
    GetNamedResultResponseGetNamedResultResponseSuccessDataChunk,
    GetNamedResultResponseGetNamedResultResponseSuccessDataSummary,
    GetJobNamedResultHeaderResponseGetJobNamedResultHeaderResponseSuccess,
    GetJobNamedResultsHeadersResponseGetJobNamedResultsHeadersResponseSuccessOutputHeader,
)


class JobResultApi(BaseApiV2[JobServiceStub]):
    def __init__(self, connection_details: ConnectionDetails, job_id: str, supports_chunk_streaming: bool):
        super().__init__(connection_details)
        self._id = job_id
        self._supports_chunk_streaming = supports_chunk_streaming

    @property
    def _stub_class(self) -> Type[JobServiceStub]:
        return JobServiceStub

    @property
    def id(self) -> str:
        return self._id

    async def get_job_named_result(
        self, output_name: str, long_offset: int, limit: int, timeout: Optional[float] = None
    ) -> AsyncIterator[JobNamedResult]:
        timeout = timeout if timeout is not None else self._timeout

        request = GetNamedResultRequest(job_id=self._id, output_name=output_name, long_offset=long_offset, limit=limit)
        if self._supports_chunk_streaming:
            results_iterator = self._run_async_iterator(self._stub.get_named_result, request, timeout=timeout)
            async for result in self._group_results(results_iterator):
                yield result
        else:
            async for response in self._stub.get_named_result(request, timeout=timeout):
                yield JobNamedResult(
                    data=response.success.data, count_of_items=response.success.count_of_items, output_name=output_name
                )

    async def get_job_named_results(
        self, data_to_fetch: Mapping[str, slice], timeout: Optional[float]
    ) -> AsyncIterator[JobNamedResult]:
        request = GetNamedResultsRequest(
            job_id=self._id,
            outputs=[
                GetNamedResultsRequestOutput(
                    output_name=output_name,
                    range=Range(from_=s.start, to=s.stop - 1),
                )
                for output_name, s in data_to_fetch.items()
            ],
        )
        timeout = timeout if timeout is not None else self._timeout
        results_iterator = self._run_async_iterator(self._stub.get_named_results, request, timeout=timeout)
        async for result in self._group_results(results_iterator):
            yield result

    async def _group_results(
        self, results_iterator: AsyncIterator[GetNamedResultResponse]
    ) -> AsyncIterator[JobNamedResult]:
        output_name_to_accumulated_bytes: Dict[str, bytes] = defaultdict(lambda: b"")
        # Note that the result name is a new attribute, so for old QOP we will receive just empty string.
        # But it is guaranteed that all the results belong to the same stream.
        async for response in results_iterator:
            _, response_val = betterproto.which_one_of(response, "response_oneof")
            if isinstance(response_val, GetNamedResultResponseGetNamedResultResponseError):
                raise DataFetchingError(f"{response_val.details}")
            if not isinstance(response_val, GetNamedResultResponseGetNamedResultResponseSuccess):
                raise ValueError(f"Unexpected response: {response}")
            name = response_val.output_name
            _, data_one_of = betterproto.which_one_of(response_val, "data_oneof")
            if isinstance(data_one_of, GetNamedResultResponseGetNamedResultResponseSuccessDataChunk):
                output_name_to_accumulated_bytes[name] += data_one_of.data
            elif isinstance(data_one_of, GetNamedResultResponseGetNamedResultResponseSuccessDataSummary):
                data = output_name_to_accumulated_bytes.pop(name)
                result = JobNamedResult(data=data, count_of_items=data_one_of.count, output_name=name)
                yield result
            else:
                raise ValueError(f"Unexpected response: {response}")

    def get_named_headers(self, name_to_flat_struct: Mapping[str, bool]) -> Mapping[str, JobNamedResultHeader]:
        request = GetJobNamedResultsHeadersRequest(
            job_id=self._id,
            outputs=[
                GetJobNamedResultsHeadersRequestOutput(output_name=name, flat_format=flat_struct)
                for name, flat_struct in name_to_flat_struct.items()
            ],
        )
        headers = self._fetch_headers(request)
        return {response.output_name: self._convert_header_response_to_model(response) for response in headers}

    def _fetch_headers(
        self, request: GetJobNamedResultsHeadersRequest
    ) -> Sequence[GetJobNamedResultsHeadersResponseGetJobNamedResultsHeadersResponseSuccessOutputHeader]:
        response = self._run(self._stub.get_job_named_results_headers(request, timeout=self._timeout))
        return response.headers

    def get_named_header(self, output_name: str, flat_struct: bool) -> JobNamedResultHeader:
        request = GetJobNamedResultHeaderRequest(job_id=self._id, output_name=output_name, flat_format=flat_struct)
        response = self._run(self._stub.get_job_named_result_header(request, timeout=self._timeout))
        return self._convert_header_response_to_model(response)

    @staticmethod
    def _convert_header_response_to_model(
        response: Union[
            GetJobNamedResultHeaderResponseGetJobNamedResultHeaderResponseSuccess,
            GetJobNamedResultsHeadersResponseGetJobNamedResultsHeadersResponseSuccessOutputHeader,
        ],
    ) -> JobNamedResultHeader:
        return JobNamedResultHeader(
            count_so_far=response.count_so_far,
            bare_dtype=response.simple_dtype,
            shape=tuple(response.shape),
            has_dataloss=response.has_data_loss,
            has_execution_errors=False,
        )

    def get_job_result_schema(self) -> Mapping[str, JobResultItemSchema]:
        request = GetJobResultSchemaRequest(job_id=self._id)
        response = self._run(self._stub.get_job_result_schema(request, timeout=self._timeout))
        return {
            item.name: JobResultItemSchema(
                name=item.name,
                bare_dtype=item.simple_dtype,
                shape=tuple(item.shape),
                is_single=item.is_single,
                expected_count=item.expected_count,
            )
            for item in response.items
        }

    def get_job_state(self) -> JobStreamingState:
        request = GetJobResultStateRequest(job_id=self._id)
        response = self._run(self._stub.get_job_result_state(request, timeout=self._timeout))
        return JobStreamingState(
            job_id=self._id,
            done=response.done,
            closed=response.closed,
            has_dataloss=response.has_dataloss,
        )

    def get_job_execution_status(self) -> JobExecutionStatus:
        request = JobServiceGetJobStatusRequest(job_id=self._id)
        response = self._run(self._stub.get_job_status(request, timeout=self._timeout))
        return response.status



import logging
import warnings
from typing import (
    Dict,
    Tuple,
    Union,
    Literal,
    Mapping,
    TypeVar,
    Callable,
    KeysView,
    Optional,
    Generator,
    ItemsView,
    Collection,
    ValuesView,
    overload,
)

from qm.utils import deprecation_message
from qm.api.v2.job_result_api import JobResultApi
from qm.api.job_result_api import JobResultServiceApi
from qm.utils.general_utils import run_until_with_timeout
from qm.type_hinting.general import NumpyArray, NumpyNumber
from qm.api.models.capabilities import QopCaps, ServerCapabilities
from qm.exceptions import JobFailedError, QmQuaException, QMTimeoutError
from qm._stream_results._multiple_streams_fetcher import MultipleStreamsFetcher
from qm._stream_results._single_stream_fetchers._single_stream_single_result_fetcher import (
    SingleStreamSingleResultFetcher,
)
from qm._stream_results._single_stream_fetchers._base_single_stream_fetcher import (
    VERY_LONG_TIME,
    AnySingleStreamFetcher,
)
from qm._stream_results._single_stream_fetchers._single_stream_multiple_results_fetcher import (
    SingleStreamMultipleResultFetcher,
)

logger = logging.getLogger(__name__)


TIMESTAMPS_LEGACY_EXT = "_timestamps"


_T = TypeVar("_T")


class StreamsManager(Mapping[str, Optional[AnySingleStreamFetcher]]):
    """Access to the results of a QmJob

    This object is created by calling [QmJob.result_handles][qm.jobs.running_qm_job.RunningQmJob.result_handles]

    Assuming you have an instance of StreamsManager:
    ```python
        job_results: StreamsManager
    ```
    This object is iterable:

    ```python
        for name, handle in job_results:
            print(name)
    ```

    Can detect if a name exists:

    ```python
    if "somename" in job_results:
        print("somename exists!")
        handle = job_results.get("somename")
    ```
    """

    def __init__(
        self,
        service: Union[JobResultServiceApi, JobResultApi],
        capabilities: ServerCapabilities,
        wait_until_func: Optional[Callable[[Literal["Done"], float], None]],
    ) -> None:
        self._service = service
        self._capabilities = capabilities
        self._wait_until_func = wait_until_func

        self._schema_items = service.get_job_result_schema()
        self._multiple_streams_fetcher: Optional[MultipleStreamsFetcher] = self._create_multiple_streams_fetcher()
        self._single_stream_fetchers = self._create_all_single_stream_fetchers()

    def _create_multiple_streams_fetcher(self) -> Optional[MultipleStreamsFetcher]:
        if self._capabilities.supports(QopCaps.multiple_streams_fetching) and isinstance(self._service, JobResultApi):
            # This MultipleStreamsFetcher implements the fetching of many streams at once, using a new API.
            return MultipleStreamsFetcher(self._schema_items, self._service)
        return None

    def _create_all_single_stream_fetchers(self) -> Mapping[str, AnySingleStreamFetcher]:
        _all_fetchers: Dict[str, AnySingleStreamFetcher] = {}

        for name, item_schema in self._schema_items.items():
            fetcher_class = (
                SingleStreamSingleResultFetcher if item_schema.is_single else SingleStreamMultipleResultFetcher
            )
            _all_fetchers[name] = fetcher_class(
                schema=item_schema,
                service=self._service,
                capabilities=self._capabilities,
                multiple_streams_fetcher=self._multiple_streams_fetcher,
            )
        return _all_fetchers

    def __len__(self) -> int:
        return len(self._single_stream_fetchers)

    def __getitem__(self, item: str) -> Optional[AnySingleStreamFetcher]:
        return self.get(item)

    def __getattr__(self, item: str) -> Optional[AnySingleStreamFetcher]:
        if item == "shape" or item == "__len__":
            return (
                None  # this is here because of a bug in pycharm debugger: ver: 2022.3.2 build #PY-223.8617.48 (24/1/23)
            )
        return self.get(item)

    def __iter__(self) -> Generator[Tuple[str, Optional[AnySingleStreamFetcher]], None, None]:  # type: ignore[override]
        warnings.warn(
            deprecation_message(
                method="streaming_result_fetcher.__iter__",
                deprecated_in="1.2.0",
                removed_in="2.0.0",
                details="This function is going to change its API to be similar to this of a dictionary, "
                "Use `iterate_results` for the old API.",
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return self.iterate_results()

    def iterate_results(self) -> Generator[Tuple[str, Optional[AnySingleStreamFetcher]], None, None]:
        for item in self._schema_items.values():
            yield item.name, self.get(item.name)

    def keys(self) -> KeysView[str]:
        """
        Returns a view of the names of the results
        """
        return self._single_stream_fetchers.keys()

    def items(self) -> ItemsView[str, AnySingleStreamFetcher]:
        """
        Returns a view, in which the first item is the name of the result and the second is the result
        """
        return self._single_stream_fetchers.items()

    def values(self) -> ValuesView[AnySingleStreamFetcher]:
        """
        Returns a view of the results
        """
        return self._single_stream_fetchers.values()

    def is_processing(self) -> bool:
        """Check if the job is still processing results

        Returns:
            True if results are still being processed, False otherwise
        """
        key = list(self._single_stream_fetchers.keys())[0]
        return self._single_stream_fetchers[key].is_processing()

    def get(
        self, name: str, /, default: Optional[Union[AnySingleStreamFetcher, _T]] = None
    ) -> Optional[Union[AnySingleStreamFetcher, _T]]:
        """Get a handle to a named result from [stream_processing][qm.qua.stream_processing]

        Args:
            name: The named result using in [stream_processing][qm.qua.stream_processing]
            default: The default value to return if the named result is unknown


        Returns:
            A handle object to the results `MultipleNamedJobResult` or `SingleNamedJobResult` or None if the named results in unknown
        """
        return self._single_stream_fetchers.get(name)

    def __contains__(self, name: object) -> bool:
        if not isinstance(name, str):
            return False
        return name in self._single_stream_fetchers

    def wait_for_all_values(self, timeout: Optional[float] = None) -> bool:
        """Wait until we know all values were processed for all named results

        Args:
            timeout: Timeout for waiting in seconds

        Returns:
            Returns True if all completed successfully; False if any result stream was closed prematurely (e.g., due to job failure or cancellation).

        """
        if self._wait_until_func is not None:
            try:
                self._wait_until_func("Done", timeout if timeout else VERY_LONG_TIME)
                return True
            except JobFailedError:
                logger.warning("Job failed or canceled, data processing has stopped, not all data is available.")
                return False
        else:

            def on_iteration() -> bool:
                all_job_states = [fetcher.get_job_state() for fetcher in self._single_stream_fetchers.values()]
                all_done = all(state.done for state in all_job_states)
                any_closed = any(state.closed for state in all_job_states)
                return all_done or any_closed

            def on_complete() -> bool:
                if all(fetcher.get_job_state().done for fetcher in self._single_stream_fetchers.values()):
                    return True
                logger.warning("Job failed or canceled, data processing has stopped, not all data is available.")
                return False

            return run_until_with_timeout(
                on_iteration_callback=on_iteration,
                on_complete_callback=on_complete,
                timeout=timeout if timeout else VERY_LONG_TIME,
                timeout_message="Job was not done in time",
            )

    @overload
    def fetch_results(
        self,
        wait_until_done: bool = True,
        timeout: float = VERY_LONG_TIME,
        stream_names: Optional[Mapping[str, Union[int, slice]]] = None,
        item: None = None,
    ) -> Mapping[str, Union[NumpyArray, Optional[NumpyNumber]]]:
        pass

    @overload
    def fetch_results(
        self,
        wait_until_done: bool = True,
        timeout: float = VERY_LONG_TIME,
        stream_names: Optional[Collection[str]] = None,
        item: Optional[Union[int, slice]] = None,
    ) -> Mapping[str, Union[NumpyArray, Optional[NumpyNumber]]]:
        pass

    def fetch_results(
        self,
        wait_until_done: bool = True,
        timeout: float = VERY_LONG_TIME,
        stream_names: Optional[Union[Mapping[str, Union[int, slice]], Collection[str]]] = None,
        item: Optional[Union[int, slice]] = None,
    ) -> Mapping[str, Union[NumpyArray, Optional[NumpyNumber]]]:
        """Fetch results from the specified streams

        Args:
            wait_until_done: If True, will wait until all results are processed before fetching
            timeout: Timeout (in seconds) that will be applied to each of the api requests.
                This means that the actual overall timeout is at least the one given, but it could be more.
            stream_names: A mapping of stream names to indices or slices to fetch, or a collection of stream names to fetch all items from
            item: An index or slice to fetch from each stream

        Returns:
            A mapping of stream names to their fetched results, represented as numpy arrays or scalars (for streams
                that return a single value)
        """
        data_to_fetch = self._standardize_fetch_args(items_to_slice=item, stream_names=stream_names)

        if wait_until_done:
            try:
                self.wait_for_all_values(timeout=timeout)
            # wait_for_all_values can return TimeoutError while our other api calls return QMTimeoutError.
            # In order to keep the error type consistent, we convert it here.
            except TimeoutError as e:
                raise QMTimeoutError(str(e)) from e

        return self._fetch_by_standard_query(data_to_fetch, timeout=timeout)

    def _fetch_by_standard_query(
        self, data_to_fetch: Mapping[str, Union[int, slice]], timeout: Optional[float] = None
    ) -> Mapping[str, Union[NumpyArray, Optional[NumpyNumber]]]:
        if self._multiple_streams_fetcher is not None:
            raw_results = self._multiple_streams_fetcher.strict_fetch(data_to_fetch, timeout=timeout)
            to_return = {}
            for name, curr_result in raw_results.items():
                to_return[name] = self._single_stream_fetchers[name]._postprocess(curr_result, flat_struct=False)
            return to_return
        else:
            results = {}
            for name, curr_item in data_to_fetch.items():
                val = self._single_stream_fetchers[name].fetch(item=curr_item, timeout=timeout)
                if val is not None:
                    results[name] = val
                else:
                    logger.warning(f"Failed to fetch results for stream '{name}'")

            return results

    def _standardize_fetch_args(
        self,
        items_to_slice: Optional[Union[int, slice]] = None,
        stream_names: Optional[Union[Mapping[str, Union[int, slice]], Collection[str]]] = None,
    ) -> Mapping[str, Union[int, slice]]:
        """Standardize the fetch arguments to a common format"""
        if isinstance(stream_names, dict) and items_to_slice is not None:
            raise QmQuaException("Cannot specify both stream_names and item")
        if stream_names is None:
            stream_names = self.keys()

        unknown_streams = set(stream_names) - set(self.keys())
        if unknown_streams:
            raise QmQuaException(f"Unknown stream names: {unknown_streams}")

        if isinstance(stream_names, dict):
            return stream_names
        else:
            items_to_slice = (
                items_to_slice if items_to_slice is not None else slice(None)
            )  # Default to fetching all items
            return {name: items_to_slice for name in stream_names}



import warnings
from typing import Union

from qm.exceptions import QmQuaException
from qm.utils import deprecation_message
from qm.qua._dsl._type_hints import MessageExpressionType
from qm.qua._expressions import literal_int, literal_bool, literal_real


def L(value: Union[bool, int, float]) -> MessageExpressionType:
    warnings.warn(deprecation_message("L", "1.2.3", "2.0.0"), DeprecationWarning, stacklevel=2)
    if isinstance(value, bool):
        return literal_bool(value)
    if isinstance(value, int):
        return literal_int(value)
    if isinstance(value, float):
        return literal_real(value)
    raise QmQuaException("literal can be bool, int or float")



import abc
from typing import Union, TypeVar, Optional

import betterproto

from qm._loc import _get_loc
from qm.type_hinting import Number
from qm.exceptions import QmQuaException
from qm.qua._expressions import QuaVariable, QuaArrayCell, QuaArrayVariable
from qm.grpc.qua import (
    QuaProgramMeasureProcess,
    QuaProgramVarRefExpression,
    QuaProgramAnalogProcessTarget,
    QuaProgramAnyScalarExpression,
    QuaProgramAnalogMeasureProcess,
    QuaProgramArrayVarRefExpression,
    QuaProgramArrayCellRefExpression,
    QuaProgramAnalogTimeDivisionSliced,
    QuaProgramIntegrationWeightReference,
    QuaProgramAnalogTimeDivisionAccumulated,
    QuaProgramAnalogTimeDivisionMovingWindow,
    QuaProgramAnalogProcessTargetTimeDivision,
    QuaProgramAnalogMeasureProcessRawTimeTagging,
    QuaProgramAnalogMeasureProcessBareIntegration,
    QuaProgramDigitalMeasureProcessRawTimeTagging,
    QuaProgramAnalogMeasureProcessDemodIntegration,
    QuaProgramAnalogMeasureProcessHighResTimeTagging,
    QuaProgramAnalogProcessTargetScalarProcessTarget,
    QuaProgramAnalogProcessTargetVectorProcessTarget,
    QuaProgramAnalogMeasureProcessDualDemodIntegration,
)


class AnalogTimeDivision(metaclass=abc.ABCMeta):
    def __init__(self, loc: str, samples_per_chunk: int):
        self.loc = loc
        self.samples_per_chunk = samples_per_chunk

    @property
    @abc.abstractmethod
    def unwrapped(self) -> QuaProgramAnalogProcessTargetTimeDivision:
        pass


class SlicedAnalogTimeDivision(AnalogTimeDivision):
    @property
    def unwrapped(self) -> QuaProgramAnalogProcessTargetTimeDivision:
        return QuaProgramAnalogProcessTargetTimeDivision(
            sliced=QuaProgramAnalogTimeDivisionSliced(samples_per_chunk=self.samples_per_chunk)
        )


class AccumulatedAnalogTimeDivision(AnalogTimeDivision):
    @property
    def unwrapped(self) -> QuaProgramAnalogProcessTargetTimeDivision:
        return QuaProgramAnalogProcessTargetTimeDivision(
            accumulated=QuaProgramAnalogTimeDivisionAccumulated(samples_per_chunk=self.samples_per_chunk)
        )


class MovingWindowAnalogTimeDivision(AnalogTimeDivision):
    def __init__(self, loc: str, samples_per_chunk: int, chunks_per_window: int):
        super().__init__(loc, samples_per_chunk)
        self.chunks_per_window = chunks_per_window

    @property
    def unwrapped(self) -> QuaProgramAnalogProcessTargetTimeDivision:
        return QuaProgramAnalogProcessTargetTimeDivision(
            moving_window=QuaProgramAnalogTimeDivisionMovingWindow(
                samples_per_chunk=self.samples_per_chunk,
                chunks_per_window=self.chunks_per_window,
            )
        )


class AnalogProcessTarget(metaclass=abc.ABCMeta):
    def __init__(self, loc: str):
        self.loc = loc

    @property
    @abc.abstractmethod
    def unwrapped(self) -> QuaProgramAnalogProcessTarget:
        pass


class ScalarProcessTarget(AnalogProcessTarget):
    def __init__(self, loc: str, target: Union[QuaVariable[float], QuaArrayCell[float]]):
        super().__init__(loc)
        self.target = target

    @property
    def unwrapped(self) -> QuaProgramAnalogProcessTarget:
        target_exp = self.target.unwrapped
        if not isinstance(target_exp, QuaProgramAnyScalarExpression):
            raise QmQuaException(f"Unknown type - {type(target_exp)}")

        target_type, found = betterproto.which_one_of(target_exp, "expression_oneof")
        if isinstance(found, QuaProgramVarRefExpression):
            target = QuaProgramAnalogProcessTargetScalarProcessTarget(variable=found)
        elif isinstance(found, QuaProgramArrayCellRefExpression):
            target = QuaProgramAnalogProcessTargetScalarProcessTarget(array_cell=found)
        else:
            raise QmQuaException(f"Unknown target type - {target_type}")
        return QuaProgramAnalogProcessTarget(scalar_process=target)


class VectorProcessTarget(AnalogProcessTarget):
    def __init__(self, loc: str, target: QuaArrayVariable[float], time_division: AnalogTimeDivision):
        super().__init__(loc)
        self.time_division = time_division
        self.target = target

    @property
    def unwrapped(self) -> QuaProgramAnalogProcessTarget:
        target_exp = self.target.unwrapped
        target = QuaProgramAnalogProcessTargetVectorProcessTarget(
            array=target_exp,
            time_division=QuaProgramAnalogProcessTargetTimeDivision().from_dict(self.time_division.unwrapped.to_dict()),
        )
        return QuaProgramAnalogProcessTarget(vector_process=target)


class MeasureProcessAbstract(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        self.loc = _get_loc()

    @property
    @abc.abstractmethod
    def unwrapped(self) -> QuaProgramMeasureProcess:
        pass


class AnalogMeasureProcess(MeasureProcessAbstract, metaclass=abc.ABCMeta):
    def __init__(self, target: AnalogProcessTarget):
        super().__init__()
        self.target = target

    @property
    def unwrapped(self) -> QuaProgramMeasureProcess:
        return QuaProgramMeasureProcess(analog=self._analog_unwrapped)

    @property
    @abc.abstractmethod
    def _analog_unwrapped(self) -> QuaProgramAnalogMeasureProcess:
        pass


class BasicIntegration(AnalogMeasureProcess):
    def __init__(self, element_output: str, iw: str, target: AnalogProcessTarget):
        super().__init__(target)
        self.element_output = element_output
        self.iw = iw

    @property
    @abc.abstractmethod
    def _analog_unwrapped(self) -> QuaProgramAnalogMeasureProcess:
        pass


class BareIntegration(BasicIntegration):
    @property
    def _analog_unwrapped(self) -> QuaProgramAnalogMeasureProcess:
        return QuaProgramAnalogMeasureProcess(
            loc=self.loc,
            bare_integration=QuaProgramAnalogMeasureProcessBareIntegration(
                integration=QuaProgramIntegrationWeightReference(name=self.iw),
                element_output=self.element_output,
                target=self.target.unwrapped,
            ),
        )


class DemodIntegration(BasicIntegration):
    @property
    def _analog_unwrapped(self) -> QuaProgramAnalogMeasureProcess:
        return QuaProgramAnalogMeasureProcess(
            loc=self.loc,
            demod_integration=QuaProgramAnalogMeasureProcessDemodIntegration(
                integration=QuaProgramIntegrationWeightReference(name=self.iw),
                element_output=self.element_output,
                target=self.target.unwrapped,
            ),
        )


class DualMeasureProcess(AnalogMeasureProcess, metaclass=abc.ABCMeta):
    def __init__(
        self,
        element_output1: str,
        element_output2: str,
        iw1: str,
        iw2: str,
        target: AnalogProcessTarget,
    ):
        super().__init__(target)
        self.element_output1 = element_output1
        self.element_output2 = element_output2
        self.iw1 = iw1
        self.iw2 = iw2


class DualDemodIntegration(DualMeasureProcess):
    @property
    def _analog_unwrapped(self) -> QuaProgramAnalogMeasureProcess:
        return QuaProgramAnalogMeasureProcess(
            loc=self.loc,
            dual_demod_integration=QuaProgramAnalogMeasureProcessDualDemodIntegration(
                integration1=QuaProgramIntegrationWeightReference(name=self.iw1),
                integration2=QuaProgramIntegrationWeightReference(name=self.iw2),
                element_output1=self.element_output1,
                element_output2=self.element_output2,
                target=self.target.unwrapped,
            ),
        )


T = TypeVar(
    "T",
    bound=Union[
        QuaProgramAnalogMeasureProcessRawTimeTagging,
        QuaProgramAnalogMeasureProcessHighResTimeTagging,
        QuaProgramDigitalMeasureProcessRawTimeTagging,
    ],
)


def _add_target_len(time_tagging: T, target_len: Optional[QuaVariable[int]]) -> T:
    if target_len is not None:
        time_tagging.target_len = target_len.unwrapped.variable
    return time_tagging


class TimeTaggingMeasurementProcess(MeasureProcessAbstract, metaclass=abc.ABCMeta):
    def __init__(
        self,
        element_output: str,
        target: QuaArrayVariable[int],
        targetLen: Optional[QuaVariable[int]],
        max_time: Number,
    ):
        super().__init__()
        self.time_tagging_target = target
        self.element_output = element_output
        self.targetLen = targetLen
        self.max_time = max_time

    @property
    def unwrapped_target(self) -> QuaProgramArrayVarRefExpression:
        a = self.time_tagging_target.unwrapped
        assert isinstance(a, QuaProgramArrayVarRefExpression)
        return a

    @property
    def unwrapped(self) -> QuaProgramMeasureProcess:
        return QuaProgramMeasureProcess(analog=self._analog_unwrapped)

    @property
    @abc.abstractmethod
    def _analog_unwrapped(self) -> QuaProgramAnalogMeasureProcess:
        pass


class RawTimeTagging(TimeTaggingMeasurementProcess):
    @property
    def _analog_unwrapped(self) -> QuaProgramAnalogMeasureProcess:
        time_tagging = QuaProgramAnalogMeasureProcessRawTimeTagging(
            max_time=int(self.max_time),
            element_output=self.element_output,
            target=self.unwrapped_target,
        )
        time_tagging = _add_target_len(time_tagging, self.targetLen)
        return QuaProgramAnalogMeasureProcess(loc=self.loc, raw_time_tagging=time_tagging)


class HighResTimeTagging(TimeTaggingMeasurementProcess):
    @property
    def _analog_unwrapped(self) -> QuaProgramAnalogMeasureProcess:
        time_tagging = QuaProgramAnalogMeasureProcessHighResTimeTagging(
            max_time=int(self.max_time),
            element_output=self.element_output,
            target=self.unwrapped_target,
        )
        time_tagging = _add_target_len(time_tagging, self.targetLen)
        return QuaProgramAnalogMeasureProcess(loc=self.loc, high_res_time_tagging=time_tagging)



import math
import time
import asyncio
import logging
import contextlib
from abc import ABCMeta, abstractmethod
from typing import Any, Type, Generic, TypeVar, Callable, Optional, Coroutine, Generator, AsyncIterator

import betterproto
import grpclib.exceptions
from grpclib import Status
from grpclib.client import Channel
from grpclib.metadata import Deadline
from grpclib.exceptions import StreamTerminatedError
from betterproto.grpc.grpclib_client import ServiceStub, MetadataLike

from qm.utils.async_utils import run_async
from qm.utils.general_utils import is_debug
from qm.api.models.server_details import ConnectionDetails
from qm.exceptions import QMTimeoutError, QMConnectionError

logger = logging.getLogger(__name__)


class GatewayNotImplementedError(NotImplementedError):
    pass


def _check_stream_terminated_is_timeout(func_start_time: float, func_timeout: float) -> bool:
    """
    When the function's timeout is reached, grpclib cancels the internally created future (distinct from the one created
    in the SDK). This cancellation prompts the server to reset the stream, triggering a "Stream reset by remote party"
    event and resulting in a StreamTerminatedError.
    This function verifies that the StreamTerminatedError is due to a timeout by ensuring that the function's
    runtime is approximately equal to the specified timeout.
    """
    func_end_time = time.time()
    func_runtime = func_end_time - func_start_time
    # Check if the function's runtime is within 10% of the specified timeout
    return math.isclose(func_runtime, func_timeout, rel_tol=0.1)


@contextlib.contextmanager
def _handle_connection_error(timeout: Optional[float] = None) -> Generator[None, None, None]:
    func_start_time = time.time()
    try:
        yield
    except grpclib.exceptions.GRPCError as e:
        if is_debug():
            logger.exception("Encountered connection error from QOP")
        if e.status == Status.UNIMPLEMENTED:
            raise GatewayNotImplementedError(
                f"Encountered connection error from QOP: details: {e.message}, status: {e.status}"
            ) from e
        raise QMConnectionError(
            f"Encountered connection error from QOP: details: {e.message}, status: {e.status}"
        ) from e
    except asyncio.TimeoutError as e:
        if is_debug():
            logger.exception("Timeout reached")
        raise QMTimeoutError("Timeout reached") from e
    except StreamTerminatedError as e:
        if is_debug():
            logger.exception("Stream terminated")

        if timeout is not None:
            if _check_stream_terminated_is_timeout(func_start_time, timeout):
                raise QMTimeoutError("Stream terminated, most likely due to a timeout") from e
        raise e


T = TypeVar("T")
StubType = TypeVar("StubType", bound=ServiceStub)
RequestMessageType = TypeVar("RequestMessageType", bound=betterproto.Message)
ResponseMessageType = TypeVar("ResponseMessageType", bound=betterproto.Message)


class BaseApi(Generic[StubType], metaclass=ABCMeta):
    def __init__(self, connection_details: ConnectionDetails):
        self._connection_details = connection_details

        self._channel = self._connection_details.channel
        self._stub: StubType = self._stub_class(self._channel)

        self._timeout: Optional[float] = self._connection_details.timeout

    def _run(
        self, coroutine: Coroutine[Any, Any, ResponseMessageType], timeout: Optional[float] = None
    ) -> ResponseMessageType:
        """
        Run a coroutine (primarily from the self._stub functions) and handle connection errors.

        Args:
            coroutine: Any coroutine function from the `self._stub` class
            timeout: A duplicate of the timeout parameter provided to the `self._stub` function. If the default
            `self._timeout` is used, this parameter can be left empty.

        Returns:
            The response message from the coroutine
        """
        if timeout is None:
            timeout = self._timeout

        with _handle_connection_error(timeout):
            return run_async(coroutine)

    async def _run_async_iterator(
        self,
        stub_func: Callable[[Any], AsyncIterator[ResponseMessageType]],
        # The parameters are typed as 'Any' because MyPy encounters issues when attempting to declare more specific types.
        request: RequestMessageType,
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> AsyncIterator[ResponseMessageType]:
        """
        Run a function that returns an AsyncIterator (primarily from the self._stub functions), and handle connection errors.

        Args:
            stub_func: A function from the `self._stub` class that returns an AsyncIterator
            request: The request message to be sent to the function
            func_kwargs: Any additional keyword arguments to be passed to the function

        Returns:
            The AsyncIterator of response messages from the function
        """

        with _handle_connection_error(timeout):
            async for response in stub_func(request, timeout=timeout, deadline=deadline, metadata=metadata):  # type: ignore[call-arg]
                yield response

    @property
    def connection_details(self) -> ConnectionDetails:
        return self._connection_details

    @property
    @abstractmethod
    def _stub_class(self) -> Type[StubType]:
        pass

    @property
    def channel(self) -> Channel:
        return self._channel
